from service_capacity_modeling.capacity_planner import planner
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern


uncertain_mid = CapacityDesires(
    service_tier=1,
    query_pattern=QueryPattern(
        estimated_read_per_second=Interval(
            low=1000, mid=10000, high=100000, confidence=0.98
        ),
        estimated_write_per_second=Interval(
            low=1000, mid=10000, high=100000, confidence=0.98
        ),
    ),
    data_shape=DataShape(
        estimated_state_size_gib=Interval(low=100, mid=500, high=1000, confidence=0.98),
    ),
)


def test_repeated_plans():
    results = []
    for _ in range(5):
        results.append(
            planner.plan(
                model_name="org.netflix.cassandra",
                region="us-east-1",
                desires=uncertain_mid,
            ).model_dump_json()
        )

    a = [hash(x) for x in results]
    # We should end up with consistent results
    assert all(i == a[0] for i in a)


def test_compositional():
    direct_result = planner.plan(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=4,
        explain=True,
    )
    composed_result = planner.plan(
        model_name="org.netflix.key-value",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=4,
        explain=True,
    )

    count = len(direct_result.least_regret)
    direct_clusters = []
    for i in range(count):
        direct_cluster = direct_result.least_regret[i].candidate_clusters.zonal[0]
        direct_clusters.append(direct_cluster)
        # FIXME(josephl): It appears that since we are now zipping the
        # regional and zonal clusters we can get repeats in the zonal.
        # This is odd to me but not related to the 6th gen instances
        composed_cluster = composed_result.least_regret[i].candidate_clusters.zonal[0]
        assert composed_cluster in direct_clusters

        java = composed_result.least_regret[i].candidate_clusters.regional[0]
        assert java.cluster_type == "dgwkv"
        # usually like 15 * 4 = ~50
        assert 100 > java.count * java.instance.cpu > 20


def test_multiple_options_diversify_with_more_simulations():
    """
    This test appears strange at first. The goal is to show that with less
    simulations we would see a smaller subset of the diverse world of outputs
    than with more simulations
    """

    # These values happen to work today but may not work in the future with
    # changes to the CP inputs (instances, costs, performance).
    # Feel free to change the numbers as long as it fits the below assertion
    arbitrary_num_results = 8
    arbitrary_small_number = 12
    arbitrary_large_number = 1024
    assert arbitrary_small_number < arbitrary_large_number

    less_simulations_result = planner.plan(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=arbitrary_num_results,
        simulations=arbitrary_small_number,
    )
    more_simulations_result = planner.plan(
        model_name="org.netflix.cassandra",
        region="us-east-1",
        desires=uncertain_mid,
        num_results=arbitrary_num_results,
        simulations=arbitrary_large_number,
    )

    # Potentially brittle assertion. This is the part likely to break
    # The idea is that we should see more options with more simulations.
    less_simulations_famlies = {
        lr.candidate_clusters.zonal[0].instance.family
        for lr in less_simulations_result.least_regret
    }
    more_simulations_families = {
        lr.candidate_clusters.zonal[0].instance.family
        for lr in more_simulations_result.least_regret
    }
    assert len(less_simulations_famlies) < len(more_simulations_families)

    expected_family_types = {"i", "r", "c", "m"}
    for f in less_simulations_famlies:
        assert f[0] in expected_family_types
    for f in more_simulations_families:
        assert f[0] in expected_family_types



# pylint: disable=too-many-lines
from __future__ import annotations

import re
import sys
from decimal import Decimal
from fractions import Fraction
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
from pydantic import BaseModel
from pydantic import computed_field
from pydantic import ConfigDict
from pydantic import Field

from service_capacity_modeling.enum_utils import enum_docstrings
from service_capacity_modeling.enum_utils import StrEnum

GIB_IN_BYTES = 1024 * 1024 * 1024
MIB_IN_BYTES = 1024 * 1024
MEGABIT_IN_BYTES = (1000 * 1000) / 8


class ExcludeUnsetModel(BaseModel):
    def model_dump(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        if "exclude_unset" not in kwargs:
            kwargs["exclude_unset"] = True
        return super().model_dump(*args, **kwargs)

    def model_dump_json(self, *args: Any, **kwargs: Any) -> str:
        if "exclude_unset" not in kwargs:
            kwargs["exclude_unset"] = True
        return super().model_dump_json(*args, **kwargs)


###############################################################################
#              Models (structs) for how we describe intervals                 #
###############################################################################


@enum_docstrings
class IntervalModel(StrEnum):
    """Statistical distribution models for approximating intervals

    When we have uncertainty intervals (low, mid, high), we need to choose
    a probability distribution to model that uncertainty for simulation purposes.
    """

    def __repr__(self) -> str:
        return f"D({self.value})"

    gamma = "gamma"
    """Gamma distribution - used for modeling right-skewed data"""

    beta = "beta"
    """Beta distribution - bounded to [0,1], good for modeling proportions
    and probabilities"""


class Interval(ExcludeUnsetModel):
    """Represents an uncertainty interval with confidence bounds"""

    low: float
    mid: float
    high: float

    confidence: float = 1.0
    """How confident are we of this interval (0.0 to 1.0)"""

    model_with: IntervalModel = IntervalModel.beta
    """How to approximate this interval for simulation (e.g. beta distribution)"""

    allow_simulate: bool = True
    """Whether to allow simulation of this interval. Some models might not support
    simulation or some properties might not want probabilistic treatment."""

    minimum_value: Optional[float] = None
    maximum_value: Optional[float] = None
    model_config = ConfigDict(frozen=True, protected_namespaces=())

    @property
    def can_simulate(self) -> bool:
        return self.confidence <= 0.99 and self.allow_simulate

    @property
    def minimum(self) -> float:
        if self.minimum_value is None:
            if self.confidence == 1.0:
                return self.low * 0.999
            return self.low / 2

        return self.minimum_value

    @property
    def maximum(self) -> float:
        if self.maximum_value is None:
            if self.confidence == 1.0:
                return self.high * 1.001
            return self.high * 2
        return self.maximum_value

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Interval):
            return False
        return self.__hash__() == other.__hash__()

    def scale(self, factor: float) -> Interval:
        minimum_value = (
            self.minimum * factor if self.minimum_value is not None else None
        )
        maximum_value = (
            self.maximum * factor if self.maximum_value is not None else None
        )
        return Interval(
            low=self.low * factor,
            mid=self.mid * factor,
            high=self.high * factor,
            confidence=self.confidence,
            model_with=self.model_with,
            allow_simulate=self.allow_simulate,
            minimum_value=minimum_value,
            maximum_value=maximum_value,
        )

    def offset(self, delta: float) -> Interval:
        minimum_value = self.minimum + delta if self.minimum_value is not None else None
        maximum_value = self.maximum + delta if self.maximum_value is not None else None
        return Interval(
            low=self.low + delta,
            mid=self.mid + delta,
            high=self.high + delta,
            confidence=self.confidence,
            model_with=self.model_with,
            allow_simulate=self.allow_simulate,
            minimum_value=minimum_value,
            maximum_value=maximum_value,
        )


class FixedInterval(Interval):
    allow_simulate: bool = False


@lru_cache(2048)
def certain_int(x: int) -> Interval:
    return Interval(low=x, mid=x, high=x, confidence=1.0)


@lru_cache(2048)
def certain_float(x: float) -> Interval:
    return Interval(low=x, mid=x, high=x, confidence=1.0)


@lru_cache(2048)
def fixed_float(x: float) -> FixedInterval:
    return FixedInterval(low=x, mid=x, high=x, confidence=1.0)


def interval(samples: Sequence[float], low_p: int = 5, high_p: int = 95) -> Interval:
    p = np.percentile(a=samples, q=[0, low_p, 50, high_p, 100])
    conf = (high_p - low_p) / 100
    return Interval(
        low=p[1],
        mid=p[2],
        high=p[3],
        minimum_value=p[0],
        maximum_value=p[4],
        confidence=conf,
    )


def normalized_aws_size(name: str) -> Fraction:
    """Normalizes an AWS shape to a fractional xlarge unit"""
    _, size = name.split(".")
    numeric = re.findall(r"\d+", size)
    if numeric:
        assert len(numeric) == 1
        return Fraction(float(numeric[0]))
    return {
        "small": Fraction(1, 8),
        "medium": Fraction(1, 4),
        "large": Fraction(1, 2),
        "xlarge": Fraction(1),
        # Is this always true?
        "metal": Fraction(48, 1),
    }[size]


###############################################################################
#              Models (structs) for how we describe hardware                  #
###############################################################################


class Lifecycle(StrEnum):
    """Represents the lifecycle of hardware from initial preview
    to end-of-life.

    For example a particular shape of hardware might be released under
    preview and adventurous models may wish to use those, and then
    more risk-averse workloads may wait for stability.

    By default models are shown beta and stable shapes to pick from
    """

    alpha = "alpha"
    beta = "beta"
    stable = "stable"
    deprecated = "deprecated"
    end_of_life = "end-of-life"


@enum_docstrings
class DriveType(StrEnum):
    """Represents the type and attachment model of storage drives

    Drives can be either local (ephemeral, instance-attached) or network-attached
    (persistent, e.g., EBS), and can use SSD or HDD storage media.
    """

    local_ssd = "local-ssd"
    """Local SSD storage - ephemeral, physically attached to instance,
    highest performance"""

    local_hdd = "local-hdd"
    """Local HDD storage - ephemeral, physically attached to instance,
    lower cost than SSD"""

    attached_ssd = "attached-ssd"
    """Network-attached SSD (e.g., EBS gp3) - persistent, survives
    instance termination"""

    attached_hdd = "attached-hhd"
    """Network-attached HDD (e.g., EBS st1) - persistent, cost-optimized
    """


class Drive(ExcludeUnsetModel):
    """Represents a cloud drive e.g. EBS or ephemeral drives

    This model is generic to any cloud
    """

    name: str
    drive_type: DriveType = DriveType.local_ssd
    size_gib: int = 0
    read_io_per_s: Optional[int] = None
    write_io_per_s: Optional[int] = None
    throughput: Optional[int] = None

    single_tenant: bool = True
    """Whether this drive has single tenant IO capacity (e.g. physical
    drive vs virtualized drive)"""

    max_scale_size_gib: int = 0
    """Maximum size this drive can scale to (in GiB)"""

    max_scale_io_per_s: int = 0
    """Maximum IOPS this drive can scale to"""

    block_size_kib: int = 4
    """Size of a single IO operation against this device (in KiB)"""

    group_size_kib: int = 4
    """When sequential, how much IO is grouped into a single operation
    (in KiB). Some cloud drives (e.g. EBS) can group sequential ops
    together and DBs take advantage.
    See: https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/
    ebs-io-characteristics.html"""

    lifecycle: Lifecycle = Lifecycle.stable
    compatible_families: List[str] = []

    annual_cost_per_gib: float = 0

    annual_cost_per_read_io: List[Tuple[float, float]] = []
    """Tiered pricing for read IOPS: list of (max_iops, annual_cost)
    tuples. Example: [(32000, 0.78), (160000, 0.384)]"""

    annual_cost_per_write_io: List[Tuple[float, float]] = []
    """Tiered pricing for write IOPS: list of (max_iops, annual_cost)
    tuples. Example for io2: [(32000, 0.78), (64000, 0.552), (256000, 0.432)]
    Note: gp3 doesn't distinguish read/write IOPS, so this would be empty for gp3."""

    read_io_latency_ms: FixedInterval = FixedInterval(
        low=0.8, mid=1, high=2, confidence=0.9
    )
    """Default read latency assumes cloud SSD (e.g. EBS gp2).
    Override in hardware description if different."""
    write_io_latency_ms: FixedInterval = FixedInterval(
        low=0.6, mid=2, high=3, confidence=0.9
    )
    """Default write latency assumes cloud SSD (e.g. EBS gp2).
    Writes are typically faster than reads due to write buffering.
    Override in hardware description if different."""

    @property
    def rand_io_size_kib(self) -> int:
        return self.block_size_kib

    @property
    def seq_io_size_kib(self) -> int:
        return max(self.block_size_kib, self.group_size_kib)

    @property
    def max_size_gib(self) -> float:
        if self.max_scale_size_gib != 0:
            return self.max_scale_size_gib
        else:
            return self.size_gib

    @property
    def max_io_per_s(self) -> int:
        if self.max_scale_io_per_s != 0:
            return self.max_scale_io_per_s
        else:
            return sys.maxsize

    @computed_field(return_type=float)  # type: ignore
    @property
    def annual_cost(self) -> float:
        size = self.size_gib or 0
        r_ios = self.read_io_per_s or 0
        w_ios = self.write_io_per_s or 0

        # Time to do income taxes ...
        # Inputs are ranges of io limits and costs for ios in that range
        # [(32000.0, 0.78),
        #  (64000.0, 0.552),
        #  (160000.0, 0.384)]
        r_cost, w_cost, offset = 0.0, 0.0, 0.0
        if self.annual_cost_per_read_io:
            for end, cost in self.annual_cost_per_read_io:
                charge_ios = min(r_ios, end) - offset
                r_cost += charge_ios * cost
                offset += charge_ios
                if offset >= r_ios:
                    break

        offset = 0.0
        if self.annual_cost_per_write_io:
            for end, cost in self.annual_cost_per_write_io:
                charge_ios = min(w_ios, end) - offset
                w_cost += charge_ios * cost
                offset += charge_ios
                if offset >= w_ios:
                    break

        return size * self.annual_cost_per_gib + r_cost + w_cost

    @staticmethod
    def get_managed_drive() -> Drive:
        return Drive(name="managed")


@enum_docstrings
class Platform(StrEnum):
    """Represents the CPU architecture or managed service platform

    Hardware can run on different CPU architectures (x86_64, ARM) or be a fully
    managed service platform (e.g., Aurora). The platform choice affects performance
    characteristics, cost, and compatibility.
    """

    amd64 = "amd64"
    """x86-64 architecture - most Intel and AMD instance types (e.g., m5,
    c5, r5)"""

    arm64 = "arm64"
    """ARM64 architecture - Graviton and other ARM-based instances (e.g.,
    m6g, c7g)"""

    aurora_mysql = "Aurora MySQL"
    """AWS Aurora MySQL-compatible managed database platform"""

    aurora_postgres = "Aurora PostgreSQL"
    """AWS Aurora PostgreSQL-compatible managed database platform"""


class Instance(ExcludeUnsetModel):
    """Represents a cloud instance aka Hardware Shape

    This model is generic to any cloud.
    """

    name: str
    cpu: int = Field(
        title="CPU threads (vCPUs or effective cores)",
        description=(
            "The number of CPU threads this instance has, typically a SMT thread "
            "but if equal to Instance.cpu_cores, actual cores."
        ),
    )
    cpu_cores: Optional[int] = Field(
        default=None,
        title="If known the number of physical cores present on the machine",
        description=(
            "Most Instance.cpu are really threads, meaning they are not truly "
            "parallel. Some Instance.cpu are full fat cores, in which case this "
            "will be set and be equal to the cpu count. Use the cores getter in code"
        ),
    )
    cpu_ghz: float = Field(
        title="The clock frequency of these cores",
        description=(
            "How fast the base clock of the cores are. We assume only baseline"
            " all-core sustained turbo for the purpose of capacity planning."
        ),
    )
    cpu_ipc_scale: float = Field(
        default=1.0,
        title="Instruction per clock scale: core speed is multiplied by this",
        description=(
            "Not all cores or ghz are created equal, and if we want to represent that "
            "a CPU is more productive per unit frequency, this fractional scaling "
            "influences core normalization logic. Note that in the longer term we "
            "expect to replace this with a standard measure of average performance."
        ),
    )
    ram_gib: float
    net_mbps: float
    drive: Optional[Drive] = None
    annual_cost: float = 0
    lifecycle: Lifecycle = Lifecycle.stable

    platforms: List[Platform] = [Platform.amd64]
    """CPU platforms this instance supports. Typically hardware has a
    single platform, but sometimes instances can run on multiple platforms
    (e.g. amd64 and arm64)."""

    family_separator: str = "."

    @property
    def family(self) -> str:
        return self.name.rsplit(self.family_separator, 1)[0]

    @property
    def size(self) -> str:
        return self.name.rsplit(self.family_separator, 1)[1]

    @property
    def cores(self) -> int:
        if self.cpu_cores is not None:
            return self.cpu_cores
        return self.cpu // 2

    @staticmethod
    def get_managed_instance() -> Instance:
        return Instance(
            name="managed.0", cpu=0, cpu_ghz=0, ram_gib=0, net_mbps=0, drive=None
        )

    def merge_with(self, overrides: "Instance") -> "Instance":
        self_dict = self.model_dump()
        other_dict = overrides.model_dump(exclude_unset=True)

        for k, v in other_dict.items():
            # TODO we need a deep merge on drive (recursive merge)
            if k in ("platforms",):
                # Unique merge platforms
                merged_platforms = list(self_dict.get("platforms", []))
                merged_platforms = merged_platforms + [
                    i
                    for i in other_dict.get("platforms", [])
                    if i not in merged_platforms
                ]
                self_dict["platforms"] = merged_platforms
            else:
                self_dict[k] = v
        return Instance(**self_dict)


default_reference_shape = Instance(
    name="default_reference_shape",
    cpu=4,
    # Much of our benchmarking was carried out in the 5th gen i2 shape, can
    # adjust this up once we go through and fix all the latency distributions
    cpu_ghz=2.3,
    cpu_ipc_scale=1.0,
    ram_gib=8.0,
    net_mbps=1000,
)


class Service(ExcludeUnsetModel):
    """Represents a cloud service, such as a blob store (S3) or
    managed service such as DynamoDB or RDS.

    This model is generic to any cloud.
    """

    name: str
    size_gib: int = 0

    annual_cost_per_gib: Union[float, List[Tuple[float, float]]] = 0
    annual_cost_per_read_io: float = 0
    annual_cost_per_write_io: float = 0
    annual_cost_per_core: float = 0

    # These defaults assume a cloud blob storage like S3
    read_io_latency_ms: FixedInterval = FixedInterval(
        low=1, mid=5, high=50, confidence=0.9
    )
    write_io_latency_ms: FixedInterval = FixedInterval(
        low=1, mid=10, high=50, confidence=0.9
    )

    def annual_cost_gib(self, data_gib: float = 0) -> float:
        if isinstance(self.annual_cost_per_gib, float):
            return self.annual_cost_per_gib * data_gib
        else:
            _annual_data = data_gib
            transfer_costs = list(self.annual_cost_per_gib)
            annual_cost = 0.0
            for transfer_cost in transfer_costs:
                if _annual_data <= 0:
                    break
                if transfer_cost[0] > 0:
                    annual_cost += (
                        min(_annual_data, transfer_cost[0]) * transfer_cost[1]
                    )
                    _annual_data -= transfer_cost[0]
                else:
                    # final remaining data transfer cost
                    annual_cost += _annual_data * transfer_cost[1]
        return annual_cost


class RegionContext(ExcludeUnsetModel):
    services: Dict[str, Service] = {}
    zones_in_region: int = 3
    num_regions: int = 3


class Hardware(ExcludeUnsetModel):
    """Represents a hardware deployment

    In EC2 this maps to:
        instances: instance type -> Instance(cpu, mem, cost, etc...)
        drives: ebs type -> Drive(cost per _GiB year_, etc...)
        services: service type -> Service(name, params, cost, etc ...)
    """

    zones_in_region: int = 3
    """Number of availability zones of compute in this region"""

    instances: Dict[str, Instance] = {}
    """Instance shapes available (e.g. instance type -> Instance with cpu,
    ram, cost, etc.)"""

    drives: Dict[str, Drive] = {}
    """Drive types available (e.g. EBS type -> Drive with cost per
    GiB/year, etc.)"""

    services: Dict[str, Service] = {}
    """Managed services available (e.g. service name -> Service with
    params, cost, etc.)"""


class GlobalHardware(ExcludeUnsetModel):
    """Represents all possible hardware shapes in all regions

    In EC2 this maps to:
        us-east-1 -> Hardware available in us-east-1
        us-east-2 -> Hardware available in us-east-2
        ...
    """

    # Per region hardware shapes
    regions: Dict[str, Hardware]


class InstancePricing(ExcludeUnsetModel):
    annual_cost: float = 0
    lifecycle: Optional[Lifecycle] = None


class DrivePricing(ExcludeUnsetModel):
    annual_cost_per_gib: float = 0
    annual_cost_per_read_io: List[Tuple[float, float]] = []
    annual_cost_per_write_io: List[Tuple[float, float]] = []


class ServicePricing(ExcludeUnsetModel):
    annual_cost_per_gib: Union[float, List[Tuple[float, float]]] = 0
    annual_cost_per_read_io: float = 0
    annual_cost_per_write_io: float = 0
    annual_cost_per_core: float = 0


class HardwarePricing(ExcludeUnsetModel):
    instances: Dict[str, InstancePricing]
    drives: Dict[str, DrivePricing]
    services: Dict[str, ServicePricing]
    zones_in_region: int = 3


class Pricing(ExcludeUnsetModel):
    regions: Dict[str, HardwarePricing]


###############################################################################
#               Models (structs) for how we plan capacity                     #
###############################################################################


@enum_docstrings
class AccessPattern(StrEnum):
    """The access pattern determines capacity planning priorities: latency-sensitive
    services target low P99 latency, while throughput-oriented services optimize
    for maximum requests per second.
    """

    latency = "latency"
    """Latency-sensitive - optimize for low P99 latency, typical for
    user-facing services"""

    throughput = "throughput"
    """Throughput-oriented - optimize for maximum RPS/bandwidth, typical
    for batch processing"""


@enum_docstrings
class AccessConsistency(StrEnum):
    """
    Generally speaking consistency is expensive, so models need to know what
    kind of consistency will be required in order to estimate CPU usage
    within a factor of 4-5x correctly.

    See: https://jepsen.io/consistency for detailed consistency model definitions.

    Ordered from weakest (cheapest) to strongest (most expensive) consistency:
    - never, best_effort, eventual, read_your_writes (single-item consistency)
    - linearizable, linearizable_stale (single-item with ordering)
    - serializable, serializable_stale (multi-item transactional)
    """

    never = "never"
    """No read guarantees - writes may never be visible (e.g.,
    fire-and-forget logging)"""

    best_effort = "best-effort"
    """Best effort - writes/reads may be lost or stale, no guarantees
    (e.g., most caches)"""

    eventual = "eventual"
    """Eventual consistency - writes eventually visible with unbounded time
    delay (e.g., DNS, S3)"""

    read_your_writes = "read-your-writes"
    """Read-your-writes - clients see their own writes immediately (e.g.,
    DynamoDB consistent reads)"""

    linearizable = "linearizable"
    """Linearizable - all operations appear atomic and in real-time order
    (e.g., etcd, Consul)"""

    linearizable_stale = "linearizable-stale"
    """Linearizable writes, stale reads allowed - writes ordered, reads may
    lag (e.g., ZooKeeper)"""

    serializable = "serializable"
    """Serializable transactions - multi-item ACID transactions (e.g.,
    CockroachDB default, Spanner)"""

    serializable_stale = "serializable-stale"
    """Serializable writes, stale reads - ACID writes with lagging reads
    (e.g., CRDB stale reads, MySQL replicas)"""


AVG_ITEM_SIZE_BYTES: int = 1024


class Consistency(ExcludeUnsetModel):
    target_consistency: Optional[AccessConsistency] = Field(
        None,
        title="Consistency requirement on access",
        description=(
            "Stronger consistency access is generally more expensive."
            " The words used here to describe consistency attempt to "
            " align with the Jepsen tree of multi/single object "
            " consistency models: https://jepsen.io/consistency"
        ),
    )
    staleness_slo_sec: FixedInterval = Field(
        FixedInterval(low=0, mid=10, high=60),
        title="When stale reads are permitted what is the staleness requirement",
        description=(
            "Eventual consistency (aka stale reads) is usually bounded by some"
            " amount of time. Applications can use this to try to enforce when "
            " a write is available for reads"
        ),
    )


class GlobalConsistency(ExcludeUnsetModel):
    same_region: Consistency = Consistency(
        target_consistency=None,
        staleness_slo_sec=FixedInterval(low=0, mid=0.1, high=1),
    )
    cross_region: Consistency = Consistency(
        target_consistency=None,
        staleness_slo_sec=FixedInterval(low=10, mid=60, high=600),
    )


class QueryPattern(ExcludeUnsetModel):
    # Will the service primarily be accessed in a latency sensitive mode
    # (aka we care about P99) or throughput (we care about averages)
    access_pattern: AccessPattern = AccessPattern.latency
    access_consistency: GlobalConsistency = GlobalConsistency()

    # A main input, how many requests per second will we handle
    # We assume this is the mean of a range of possible outcomes
    estimated_read_per_second: Interval = certain_int(0)
    estimated_write_per_second: Interval = certain_int(0)

    # A main input, how much _on cpu_ time per operation do you take.
    # This depends heavily on workload, but this is a generally ok default
    # For a Java app (C or C++ will generally be about 10x better,
    # python 2-4x slower, etc...)
    estimated_mean_read_latency_ms: Interval = certain_float(1)
    estimated_mean_write_latency_ms: Interval = certain_float(1)

    # For stateful services the amount of data accessed per
    # read and write impacts disk and network provisioniong
    # For stateless services it mostly just impacts memory and network
    estimated_mean_read_size_bytes: Interval = certain_int(AVG_ITEM_SIZE_BYTES)
    estimated_mean_write_size_bytes: Interval = certain_int(AVG_ITEM_SIZE_BYTES // 2)

    # For workloads which have bursts of async work, what is the
    # expected parallelism of those workloads. Note the summation of
    # read and write parallelism will lower bound the number of cores.
    estimated_read_parallelism: Interval = Field(
        certain_int(1),
        title="Estimated per instance parallelism on read operations",
        description=(
            "The estimated amount of parallel work streams on a single "
            "host. For example a read triggers async callbacks that need "
            "to be executed truly in parallel (not just concurrent)."
        ),
    )
    estimated_write_parallelism: Interval = Field(
        certain_int(1),
        title="Estimated per instance parallelism on write operations",
        description=(
            "The estimated amount of parallel work streams on a single "
            "host. For example a write triggers async fanouts that need "
            "to be executed truly in parallel (not just concurrent)."
        ),
    )

    # The latencies at which oncall engineers get involved. We want
    # to provision such that we don't involve oncall
    # Note that these summary statistics will be used to create reasonable
    # distribution approximations of these operations (yielding p25, p99, etc)
    read_latency_slo_ms: FixedInterval = FixedInterval(
        low=0.4, mid=4, high=10, confidence=0.98
    )
    write_latency_slo_ms: FixedInterval = FixedInterval(
        low=0.4, mid=4, high=10, confidence=0.98
    )


class DataShape(ExcludeUnsetModel):
    estimated_state_size_gib: Interval = Field(
        certain_int(0),
        title="Estimated amount of state in the system in GiB",
        description=(
            "The estimated amount of state that will be stored."
            " Note that this is an estimate and doesn't need to be exact"
        ),
    )
    estimated_state_item_count: Optional[Interval] = None

    estimated_working_set_percent: Optional[Interval] = Field(
        None,
        title="Estimated working set percentage",
        description=(
            "The estimated percentage of data that will be accessed frequently"
            " and therefore must be kept hot in memory (e.g. 0.10). Note that "
            " models will generally estimate this from the latency SLO and "
            "latency model of the drives being attached"
        ),
    )

    estimated_compression_ratio: Interval = certain_float(1)
    """Data compression ratio (forward ratio). Examples:
    - 2 means 2:1 compression (0.5x on-disk size)
    - 5 means 5:1 compression (0.2x on-disk size)
    Note: databases may have different compression strategies affecting this."""

    reserved_instance_app_mem_gib: float = 2
    """Fixed memory per instance for application (e.g. process heap memory)"""

    reserved_instance_system_mem_gib: float = 1
    """Fixed memory per instance for system (e.g. kernel and system processes)"""

    # How durable does this dataset need to be. We want to provision
    # sufficient replication and backups of data to achieve the target
    # durability SLO so we don't lose our customer's data. Note that
    # This is measured in orders of magnitude. So
    #   1000   = 1 - (1/1000) = 0.999
    #   10000  = 1 - (1/10000) = 0.9999
    durability_slo_order: FixedInterval = FixedInterval(
        low=1000, mid=10000, high=100000, confidence=0.98
    )


class CurrentClusterCapacity(ExcludeUnsetModel):
    cluster_instance_name: str
    cluster_instance: Optional[Instance] = None
    cluster_drive: Optional[Drive] = None
    cluster_instance_count: Interval
    # The distribution cpu utilization in the cluster.
    cpu_utilization: Interval = certain_float(0.0)
    # The per node distribution of memory used in gib.
    memory_utilization_gib: Interval = certain_float(0.0)
    # The per node distribution of network used in mbps.
    network_utilization_mbps: Interval = certain_float(0.0)
    # The per node distribution of disk used in gib.
    disk_utilization_gib: Interval = certain_float(0.0)


# For services that are provisioned by zone (e.g. Cassandra, EVCache)
class CurrentZoneClusterCapacity(CurrentClusterCapacity):
    pass


# For services that are provisioned regionally (e.g. Java services, RDS, etc ..)
class CurrentRegionClusterCapacity(CurrentClusterCapacity):
    pass


class CurrentClusters(ExcludeUnsetModel):
    zonal: Sequence[CurrentZoneClusterCapacity] = []
    regional: Sequence[CurrentRegionClusterCapacity] = []
    services: Sequence[ServiceCapacity] = []


@enum_docstrings
class BufferComponent(StrEnum):
    """Represents well known buffer components such as compute and storage

    Note that while these are common and defined here for models to share,
    models can have w.e. buffers they want, and this type should not enter
    the Buffers interface itself (should be str).
    """

    compute = "compute"
    """[Query Pattern] a.k.a. "Traffic" related buffers, e.g. CPU and Network"""

    storage = "storage"
    """[Data Shape] a.k.a. "Dataset" related buffers, e.g. Disk and Memory"""

    cpu = "cpu"
    """Resource specific component - CPU"""

    network = "network"
    """Resource specific component - Network"""

    disk = "disk"
    """Resource specific component - Disk"""

    memory = "memory"
    """Resource specific component - Memory"""

    @staticmethod
    def is_generic(component: str) -> bool:
        return component in {BufferComponent.compute, BufferComponent.storage}

    @staticmethod
    def is_specific(component: str) -> bool:
        return not BufferComponent.is_generic(component)


@enum_docstrings
class BufferIntent(StrEnum):
    """Defines the intent of buffer directives for capacity planning"""

    desired = "desired"
    """Most buffers show "desired" buffer, this is the default"""

    scale = "scale"
    """Ratio on top of existing buffers to ensure exists. Generally combined with a
    different desired buffer to ensure we don't just scale needlessly. This means we
    can scale up or down as long as we meet the desired buffer."""

    preserve = "preserve"
    """DEPRECATED - Use scale_up/scale_down instead. Ignores model preferences, just
    preserve existing buffers. We rarely actually want to do this since it can cause
    severe over provisioning."""

    scale_up = "scale_up"
    """Scale up if necessary to meet the desired buffer. If the existing resource is
    over-provisioned, do not reduce the requirement. If under-provisioned, the
    requirement can be increased to meet the desired buffer.

    Example: need 20 cores but have 10 → scale up to 20 cores.

    Example 2: need 20 cores but have 40 → do not scale down and require
    at least 40 cores."""

    scale_down = "scale_down"
    """Scale down if necessary to meet the desired buffer. If the existing resource is
    under-provisioned, do not increase the requirement. If over-provisioned, the
    requirement can be decreased to meet the desired buffer.

    Example: need 20 cores but have 10 → maintain buffer and do not scale up.

    Example 2: need 20 cores but have 40 → scale down to 20 cores."""


class Buffer(ExcludeUnsetModel):
    """Represents a buffer (headroom) directive for capacity planning"""

    ratio: float = 1.0
    """The buffer value expressed as a ratio over normal load (e.g. 1.5 =
    50% headroom)"""

    intent: BufferIntent = BufferIntent.desired
    """The intent of this buffer directive (almost always 'desired')"""

    components: List[str] = [BufferComponent.compute]
    """The capacity components this buffer influences (almost always
    'compute' for IPC success)"""

    sources: Dict[str, Buffer] = {}
    """If this buffer was composed from other buffers, what
    contributed"""

    explanation: str = ""
    """Optional context for why this buffer exists (e.g. 'background
    processing', 'bursty workload')"""


class Buffers(ExcludeUnsetModel):
    """Typical buffers (headroom) over the requirements to build into the system

    Note that typically callers make buffer choices based on business context for
    example a tier 1 service may request:

    Buffers(
        desired={
            "compute": Buffer(ratio: 1.5),
        }
    )
    And then models layer in their buffers, for example if a workload
    requires 10 CPU cores, but the operator of that workload  likes to build in
    2x buffer for background work (20 cores provisioned), they would express that
    as a model desire default of:

    Buffers(
        desired={
            "background": Buffer(ratio: 2.0, components=[BufferComponent.cpu]),
        }
    )

    In this case when we query the buffer_for_components(components=["cpu"])
    we get 2.0 * 1.5 = 3x because the model has reserved 2x for background work
    and the caller has asked for 1.5x load buffer - we also see both component
    buffers in the Buffer.source field.
    """

    # The default buffer if a specific buffer isn't known
    # Models should prefer to document their precise buffers in desired
    default: Buffer = Buffer(ratio=1.5)
    # Desired compute, storage, cpu, memory, etc... buffers
    desired: Dict[str, Buffer] = {}
    # Derive these buffers from current clusters or model context
    # Buffer.intent MUST be set on these Buffers to something other than "desired":
    #   scale    = ratio on top of existing buffers to ensure. Let the "derived"
    #              buffer multiplied by this ratio "needed". If the "needed" buffer
    #              is greater than desired, the needed buffer is created. If the
    #              "needed" buffer is less than desired, the needed buffer is created.
    #   preserve = ignore desired buffer entirely, just maintain existing buffers
    derived: Dict[str, Buffer] = {}


class CapacityDesires(ExcludeUnsetModel):
    # How critical is this cluster, impacts how much "extra" we provision
    # 0 = Critical to the product            (Product does not function)
    # 1 = Important to product with fallback (User experience degraded)
    # 2 = Care about it but don't wake up    (Internal apps)
    # 3 = Do not care                        (Testing)
    service_tier: int = 1

    # How will the service be queried
    query_pattern: QueryPattern = QueryPattern()

    # What will the state look like that is being queries
    data_shape: DataShape = DataShape()

    # What is the current deployment and utilization of the system
    current_clusters: Optional[CurrentClusters] = None

    # What are the desired buffers (headroom) state, mostly injected by models
    # Note if you pass current_clusters you can express "I need the status quo"
    # by setting the buffers you want to preserve in buffers.derived
    buffers: Buffers = Buffers()

    @property
    def reference_shape(self) -> Instance:
        if not self.current_clusters:
            return default_reference_shape

        zonal, regional = (self.current_clusters.zonal, self.current_clusters.regional)
        if zonal and regional:
            raise ValueError(
                "The current cluster should not have both "
                "zonal and regional instances. They're mutually exclusive."
            )

        if zonal and zonal[0].cluster_instance:
            return zonal[0].cluster_instance

        if regional and regional[0].cluster_instance:
            return regional[0].cluster_instance

        return default_reference_shape

    def merge_with(self, defaults: "CapacityDesires") -> "CapacityDesires":
        # Now merge with the models default
        desires_dict = self.model_dump()
        default_dict = defaults.model_dump()

        default_dict.get("query_pattern", {}).update(
            desires_dict.pop("query_pattern", {})
        )
        default_dict.get("data_shape", {}).update(desires_dict.pop("data_shape", {}))

        # Buffers has deep structure we want to deep merge on
        if "buffers" not in default_dict:
            default_dict["buffers"] = {}
        default_buffers = default_dict["buffers"]
        desired_buffers = desires_dict.pop("buffers", {})
        if "default" in desired_buffers:
            default_buffers["default"] = desired_buffers["default"]
        for k, v in desired_buffers.get("desired", {}).items():
            default_buffers["desired"][k] = v

        default_buffers.setdefault("derived", {})
        for k, v in desired_buffers.get("derived", {}).items():
            default_buffers["derived"][k] = v

        default_dict.update(desires_dict)

        desires = CapacityDesires(**default_dict)

        # If user gave state item count but not size or size but not count
        # calculate the missing one from the other
        user_size = (
            self.model_dump()
            .get("data_shape", {})
            .get("estimated_state_size_gib", None)
        )
        user_count = self.data_shape.estimated_state_item_count
        item_size_bytes = desires.query_pattern.estimated_mean_write_size_bytes.mid
        if user_size is None and user_count is not None:
            desires.data_shape.estimated_state_size_gib = user_count.scale(
                factor=(item_size_bytes / GIB_IN_BYTES)
            )
        elif user_size is not None and user_count is None:
            user_size_gib = self.data_shape.estimated_state_size_gib
            desires.data_shape.estimated_state_item_count = user_size_gib.scale(
                factor=(GIB_IN_BYTES / item_size_bytes)
            )

        return desires


class CapacityRequirement(ExcludeUnsetModel):
    requirement_type: str

    # cpu_cores was calculated relative to this reference, mostly for
    # comparing latency to clock frequency
    reference_shape: Instance = default_reference_shape
    cpu_cores: Interval
    mem_gib: Interval = certain_int(0)
    network_mbps: Interval = certain_int(0)
    disk_gib: Interval = certain_int(0)

    context: Dict[str, Any] = {}


class ClusterCapacity(ExcludeUnsetModel):
    cluster_type: str

    count: int
    instance: Instance
    attached_drives: Sequence[Drive] = ()
    annual_cost: float
    # When provisioning services we might need to signal they
    # should have certain configuration, for example flags that
    # affect durability shut off
    cluster_params: Dict[str, Any] = {}


class ServiceCapacity(ExcludeUnsetModel):
    service_type: str
    annual_cost: float
    # If this cost should impact regret. Usually they do not since they
    # are proportional to the input desire not the output cluster
    regret_cost: bool = False
    # Often while provisioning cloud services we need to represent
    # parameters to the cloud APIs, use this to inject those from models
    service_params: Dict[str, Any] = {}


# For services that are provisioned by zone (e.g. Cassandra, EVCache)
class ZoneClusterCapacity(ClusterCapacity):
    pass


# For services that are provisioned regionally (e.g. Java services, RDS, etc ..)
class RegionClusterCapacity(ClusterCapacity):
    pass


class Requirements(ExcludeUnsetModel):
    zonal: Sequence[CapacityRequirement] = []
    regional: Sequence[CapacityRequirement] = []

    # Commonly a model regrets "spend", lack of "disk" space and sometimes
    # lack of "mem"ory. Default options are ["cost", "disk", "mem"]
    regrets: Sequence[str] = ("spend", "disk")

    # Used by models to have custom regret components
    # pylint: disable=unused-argument
    @staticmethod
    def regret(
        name: str, optimal_plan: "CapacityPlan", proposed_plan: "CapacityPlan"
    ) -> float:
        return 0.0


class Clusters(ExcludeUnsetModel):
    annual_costs: Dict[str, Decimal]
    zonal: Sequence[ZoneClusterCapacity] = []
    regional: Sequence[RegionClusterCapacity] = []
    services: Sequence[ServiceCapacity] = []

    # Backwards compatibility for total_annual_cost
    @computed_field(return_type=float)  # type: ignore
    @property
    def total_annual_cost(self) -> float:
        return round(float(sum(self.annual_costs.values())), 2)


class CapacityPlan(ExcludeUnsetModel):
    requirements: Requirements
    candidate_clusters: Clusters
    rank: int = 0


# Parameters to cost functions of the form
# let y = the optimal value
# let x = the proposed value
# let cost = (x - y) ^ exponent
class Regret(ExcludeUnsetModel):
    over_provision_cost: float = 0
    under_provision_cost: float = 0
    exponent: float = 1.0


class CapacityRegretParameters(ExcludeUnsetModel):
    # How much do we regret spending too much or too little money
    spend: Regret = Regret(
        over_provision_cost=1, under_provision_cost=1.25, exponent=1.2
    )

    # For every GiB we are underprovisioned by default cost $1 / year / GiB
    disk: Regret = Regret(
        over_provision_cost=0, under_provision_cost=1.1, exponent=1.05
    )

    # For every GiB we are underprovisioned on memory (for datastores
    # storing data in RAM), regret under_provisioning slightly more than disk
    mem: Regret = Regret(over_provision_cost=0, under_provision_cost=1.5, exponent=1.1)

    # Any additional metric we want to regret from the models
    # just have to be returned in the requirement
    extra: Dict[str, Regret] = {}


class PlanExplanation(ExcludeUnsetModel):
    regret_params: CapacityRegretParameters
    regret_clusters_by_model: Dict[
        str, Sequence[Tuple[CapacityPlan, CapacityDesires, float]]
    ] = {}
    desires_by_model: Dict[str, CapacityDesires] = {}
    context: Dict[str, Any] = {}


class UncertainCapacityPlan(ExcludeUnsetModel):
    requirements: Requirements
    least_regret: Sequence[CapacityPlan]
    mean: Sequence[CapacityPlan]
    percentiles: Dict[int, Sequence[CapacityPlan]]
    explanation: PlanExplanation



import logging
import math
from decimal import Decimal
from typing import Any
from typing import Dict
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import Buffer
from service_capacity_modeling.interface import BufferComponent
from service_capacity_modeling.interface import Buffers
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import ZoneClusterCapacity
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import buffer_for_components
from service_capacity_modeling.models.common import compute_stateful_zone
from service_capacity_modeling.models.common import get_effective_disk_per_node_gib
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import simple_network_mbps
from service_capacity_modeling.models.common import sqrt_staffed_cores
from service_capacity_modeling.models.common import working_set_from_drive_and_slo
from service_capacity_modeling.stats import dist_for_interval

logger = logging.getLogger(__name__)


# Pebble does Leveled compaction with tieres of size??
# (FIXME) What does pebble actually do
def _crdb_io_per_read(node_size_gib: float, sstable_size_mb: int = 1000) -> int:
    gb = node_size_gib * 1024
    sstables = max(1, gb // sstable_size_mb)
    # 10 sstables per level, plus 1 for L0 (avg)
    levels = 1 + int(math.ceil(math.log(sstables, 10)))
    return levels


def _estimate_cockroachdb_requirement(  # noqa=E501 pylint: disable=too-many-positional-arguments
    instance: Instance,
    desires: CapacityDesires,
    working_set: float,
    reads_per_second: float,
    max_rps_to_disk: int,
    zones_per_region: int = 3,
    copies_per_region: int = 3,
) -> CapacityRequirement:
    """Estimate the capacity required for one zone given a regional desire

    The input desires should be the **regional** desire, and this function will
    return the zonal capacity requirement
    """
    needed_cores = normalize_cores(
        # Keep half of the cores free for background work (compaction, backup, index)
        core_count=sqrt_staffed_cores(desires) * 2,
        target_shape=instance,
        reference_shape=desires.reference_shape,
    )
    # Keep half of the bandwidth available for backup
    needed_network_mbps = simple_network_mbps(desires) * 2

    needed_disk = math.ceil(
        (1.0 / desires.data_shape.estimated_compression_ratio.mid)
        * desires.data_shape.estimated_state_size_gib.mid
        * copies_per_region
    )

    # Rough estimate of how many instances we would need just for the the CPU
    # Note that this is a lower bound, we might end up with more.
    rough_count = math.ceil(needed_cores / instance.cpu)

    # Generally speaking we want fewer than some number of reads per second
    # hitting disk per instance. If we don't have many reads we don't need to
    # hold much data in memory.
    instance_rps = max(1, reads_per_second // rough_count)
    disk_rps = instance_rps * _crdb_io_per_read(max(1, needed_disk // rough_count))
    rps_working_set = min(1.0, disk_rps / max_rps_to_disk)

    # If disk RPS will be smaller than our target because there are no
    # reads, we don't need to hold as much data in memory
    needed_memory = min(working_set, rps_working_set) * needed_disk

    # Now convert to per zone
    needed_cores = max(1, needed_cores // zones_per_region)
    needed_disk = max(1, needed_disk // zones_per_region)
    needed_memory = max(1, int(needed_memory // zones_per_region))
    logger.debug(
        "Need (cpu, mem, disk, working) = (%s, %s, %s, %f)",
        needed_cores,
        needed_memory,
        needed_disk,
        working_set,
    )

    return CapacityRequirement(
        requirement_type="crdb-zonal",
        reference_shape=desires.reference_shape,
        cpu_cores=certain_int(needed_cores),
        mem_gib=certain_float(needed_memory),
        disk_gib=certain_float(needed_disk),
        network_mbps=certain_float(needed_network_mbps),
        context={
            "working_set": min(working_set, rps_working_set),
            "rps_working_set": rps_working_set,
            "disk_slo_working_set": working_set,
            "replication_factor": copies_per_region,
            "compression_ratio": round(
                1.0 / desires.data_shape.estimated_compression_ratio.mid, 2
            ),
            "read_per_second": reads_per_second,
        },
    )


def _upsert_params(cluster: ZoneClusterCapacity, params: Dict[str, Any]) -> None:
    if cluster.cluster_params:
        cluster.cluster_params.update(params)
    else:
        cluster.cluster_params = params


# pylint: disable=too-many-locals
def _estimate_cockroachdb_cluster_zonal(  # noqa=E501 pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    desires: CapacityDesires,
    zones_per_region: int = 3,
    copies_per_region: int = 3,
    max_local_data_per_node_gib: int = 2048,
    max_regional_size: int = 288,
    max_rps_to_disk: int = 500,
    min_vcpu_per_instance: int = 4,
    license_fee_per_core: float = 0.0,
) -> Optional[CapacityPlan]:
    if instance.cpu < min_vcpu_per_instance:
        return None

    # Right now CRDB doesn't deploy to cloud drives, just adding this
    # here and leaving the capability to handle cloud drives for the future
    if instance.drive is None:
        return None

    rps = desires.query_pattern.estimated_read_per_second.mid // zones_per_region

    # Based on the disk latency and the read latency SLOs we adjust our
    # working set to keep more or less data in RAM. Faster drives need
    # less fronting RAM.
    ws_drive = instance.drive or drive
    working_set = working_set_from_drive_and_slo(
        drive_read_latency_dist=dist_for_interval(ws_drive.read_io_latency_ms),
        read_slo_latency_dist=dist_for_interval(
            desires.query_pattern.read_latency_slo_ms
        ),
        estimated_working_set=desires.data_shape.estimated_working_set_percent,
        # CRDB has looser latency SLOs but we still want a lot of the data
        # hot in cache. Target the 95th percentile of disk latency to
        # keep in RAM.
        target_percentile=0.95,
    ).mid

    requirement = _estimate_cockroachdb_requirement(
        instance=instance,
        desires=desires,
        working_set=working_set,
        reads_per_second=rps,
        zones_per_region=zones_per_region,
        copies_per_region=copies_per_region,
        max_rps_to_disk=max_rps_to_disk,
    )

    base_mem = (
        desires.data_shape.reserved_instance_app_mem_gib
        + desires.data_shape.reserved_instance_system_mem_gib
    )

    disk_buffer_ratio = buffer_for_components(
        buffers=desires.buffers, components=[BufferComponent.disk]
    ).ratio
    max_data_per_node_gib = get_effective_disk_per_node_gib(
        instance,
        drive,
        disk_buffer_ratio,
        max_local_data_per_node_gib=max_local_data_per_node_gib,
    )
    needed_disk_gib = requirement.disk_gib.mid * disk_buffer_ratio
    min_count = math.ceil(needed_disk_gib / max_data_per_node_gib)

    cluster = compute_stateful_zone(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=needed_disk_gib,
        needed_memory_gib=requirement.mem_gib.mid,
        needed_network_mbps=requirement.network_mbps.mid,
        # Take into account the reads per read
        # from the per node dataset using leveled compaction
        # FIXME: I feel like this can be improved
        required_disk_ios=lambda size, count: (
            _crdb_io_per_read(size) * math.ceil(rps / count),
            # TODO: presumably there are some write IOs here
            0,
        ),
        # cockroachdb clusters will autobalance across available nodes
        cluster_size=lambda x: x,
        min_count=min_count,
        # Sidecars/System takes away memory from cockroachdb
        # cockroachdb by default uses --max-sql-memory of 25% of system memory
        # that cannot be used for caching
        reserve_memory=lambda x: base_mem + 0.25 * x,
        # TODO: Figure out how much memory CRDB needs to buffer writes
        # in memtables in order to only flush occasionally
        # write_buffer=...
        # required_write_buffer_gib=...
    )

    # Communicate to the actual provision that if we want reduced RF
    params = {"cockroachdb.copies": copies_per_region}
    _upsert_params(cluster, params)

    # cockroachdb clusters generally should try to stay under some total number
    # of nodes. Orgs do this for all kinds of reasons such as
    #   * Security group limits. Since you must have < 500 rules if you're
    #       ingressing public ips)
    #   * Maintenance. If your restart script does one node at a time you want
    #       smaller clusters so your restarts don't take months.
    #   * NxN network issues. Sometimes smaller clusters of bigger nodes
    #       are better for network propagation
    if cluster.count > (max_regional_size // zones_per_region):
        return None

    ec2_cost = zones_per_region * cluster.annual_cost
    license_fee = zones_per_region * (cluster.instance.cpu * license_fee_per_core)

    cluster.cluster_type = "cockroachdb"
    clusters = Clusters(
        annual_costs={"cockroachdb-zonal": Decimal(ec2_cost + license_fee)},
        zonal=[cluster] * zones_per_region,
        regional=[],
    )

    return CapacityPlan(
        requirements=Requirements(zonal=[requirement] * zones_per_region),
        candidate_clusters=clusters,
    )


class NflxCockroachDBArguments(BaseModel):
    copies_per_region: int = Field(
        default=3,
        description="How many copies of the data will exist e.g. RF=3. If unsupplied"
        " this will be deduced from durability and consistency desires",
    )
    max_regional_size: int = Field(
        default=288,
        description="What is the maximum size of a cluster in this region",
    )
    max_local_disk_gib: int = Field(
        default=2048,
        description="The maximum amount of data we store per machine",
    )
    max_rps_to_disk: int = Field(
        default=500,
        description="How many disk IOs should be allowed to hit disk per instance",
    )


class NflxCockroachDBCapacityModel(CapacityModel):
    @staticmethod
    def default_buffers() -> Buffers:
        return Buffers(
            default=Buffer(ratio=1.2),
        )

    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # (FIXME): Need crdb input
        # TODO: Use read requirements to compute RF.
        copies_per_region: int = extra_model_arguments.get("copies_per_region", 3)
        max_regional_size: int = extra_model_arguments.get("max_regional_size", 500)
        max_rps_to_disk: int = extra_model_arguments.get("max_rps_to_disk", 500)
        # Very large nodes are hard to recover
        max_local_data_per_node_gib: int = extra_model_arguments.get(
            "max_local_data_per_node_gib",
            extra_model_arguments.get("max_local_disk_gib", 2048),
        )

        # Cockroach Labs recommends a minimum of 8 vCPUs and strongly
        # recommends no fewer than 4 vCPUs per node.
        min_vcpu_per_instance: int = extra_model_arguments.get(
            "min_vcpu_per_instance", 4
        )
        license_fee_per_core: float = context.services[
            "crdb_core_license"
        ].annual_cost_per_core

        return _estimate_cockroachdb_cluster_zonal(
            instance=instance,
            drive=drive,
            desires=desires,
            zones_per_region=context.zones_in_region,
            copies_per_region=copies_per_region,
            max_regional_size=max_regional_size,
            max_local_data_per_node_gib=max_local_data_per_node_gib,
            max_rps_to_disk=max_rps_to_disk,
            min_vcpu_per_instance=min_vcpu_per_instance,
            license_fee_per_core=license_fee_per_core,
        )

    @staticmethod
    def description() -> str:
        return "Netflix Streaming CockroachDB Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxCockroachDBArguments.model_json_schema()

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        acceptable_consistency = {
            None,
            AccessConsistency.linearizable,
            AccessConsistency.linearizable_stale,
            AccessConsistency.serializable,
            AccessConsistency.serializable_stale,
            AccessConsistency.never,
        }
        for key, value in user_desires.query_pattern.access_consistency:
            if value.target_consistency not in acceptable_consistency:
                raise ValueError(
                    f"CockroachDB can only provide {acceptable_consistency} access."
                    f"User asked for {key}={value}"
                )

        buffers = NflxCockroachDBCapacityModel.default_buffers()
        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.serializable,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.serializable,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    # (FIXME): Need crdb input
                    # CockroachDB reads and writes can take CPU time as
                    # JOINs and such can be hard to predict.
                    estimated_mean_read_latency_ms=Interval(
                        low=1, mid=2, high=100, confidence=0.98
                    ),
                    # Writes typically involve transactions which can be
                    # expensive, but it's rare for it to have a huge tail
                    estimated_mean_write_latency_ms=Interval(
                        low=1, mid=4, high=200, confidence=0.98
                    ),
                    # Assume point queries "Single digit millisecond SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=1,
                        mid=10,
                        high=20,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=1,
                        mid=10,
                        high=20,
                        confidence=0.98,
                    ),
                ),
                # Most latency sensitive cockroachdb clusters are in the
                # < 100GiB range
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=20, high=100, confidence=0.98
                    ),
                    # Pebble compresses with Snappy by default
                    estimated_compression_ratio=Interval(
                        minimum_value=1.1,
                        maximum_value=8,
                        low=1.5,
                        mid=2.4,
                        high=4,
                        confidence=0.98,
                    ),
                    # CRDB doesn't have a sidecar, but it does have data
                    # gateway taking about 1 MiB of memory
                    reserved_instance_app_mem_gib=0.001,
                ),
                buffers=buffers,
            )
        else:
            return CapacityDesires(
                # (FIXME): Need to pair with crdb folks on the exact values
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.throughput,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.serializable,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.serializable,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=4096, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=128, mid=4096, high=65536, confidence=0.95
                    ),
                    # (FIXME): Need crdb input
                    # CockroachDB analytics reads probably take extra time
                    # as they are full table scanning or doing complex JOINs
                    estimated_mean_read_latency_ms=Interval(
                        low=1, mid=20, high=100, confidence=0.98
                    ),
                    # Throughput writes typically involve large transactions
                    # which can be expensive, but it's rare for it to have a
                    # huge tail
                    estimated_mean_write_latency_ms=Interval(
                        low=1, mid=20, high=100, confidence=0.98
                    ),
                    # Assume scan queries "Tens of millisecond SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=10,
                        mid=20,
                        high=99,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=10,
                        mid=20,
                        high=99,
                        confidence=0.98,
                    ),
                ),
                # Most throughput sensitive cockroachdb clusters are in the
                # < 100GiB range
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=20, high=100, confidence=0.98
                    ),
                    # Pebble compresses with Snappy by default
                    estimated_compression_ratio=Interval(
                        minimum_value=1.1,
                        maximum_value=8,
                        low=1.5,
                        mid=2.4,
                        high=4,
                        confidence=0.98,
                    ),
                    # CRDB doesn't have a sidecar, but it does have data
                    # gateway taking about 1 MiB of memory
                    reserved_instance_app_mem_gib=0.001,
                ),
                buffers=buffers,
            )


nflx_cockroachdb_capacity_model = NflxCockroachDBCapacityModel()



from service_capacity_modeling.models.org.netflix import models


def test_model_arguments():
    for model in models().values():
        schema = model.extra_model_arguments_schema()
        assert schema is not None



import logging
import math
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Set

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import Buffer
from service_capacity_modeling.interface import BufferComponent
from service_capacity_modeling.interface import Buffers
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import CurrentClusterCapacity
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.interface import ServiceCapacity
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import buffer_for_components
from service_capacity_modeling.models.common import compute_stateful_zone
from service_capacity_modeling.models.common import DerivedBuffers
from service_capacity_modeling.models.common import get_effective_disk_per_node_gib
from service_capacity_modeling.models.common import network_services
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import simple_network_mbps
from service_capacity_modeling.models.common import sqrt_staffed_cores
from service_capacity_modeling.models.common import working_set_from_drive_and_slo
from service_capacity_modeling.models.common import zonal_requirements_from_current
from service_capacity_modeling.models.utils import is_power_of_2
from service_capacity_modeling.models.utils import next_doubling
from service_capacity_modeling.models.utils import next_power_of_2
from service_capacity_modeling.stats import dist_for_interval

logger = logging.getLogger(__name__)

BACKGROUND_BUFFER = "background"
CRITICAL_TIERS: Set[int] = {0, 1}
# cluster size aka nodes per ASG
CRITICAL_TIER_MIN_CLUSTER_SIZE = 2


def _write_buffer_gib_zone(
    desires: CapacityDesires, zones_per_region: int, flushes_before_compaction: int = 4
) -> float:
    # Cassandra has to buffer writes before flushing to disk, and assuming
    # we will compact every 4 flushes and we want no more than 2 redundant
    # compactions in an hour, we want <= 4**2 = 16 flushes per hour
    # or a flush of data every 3600 / 16 = 225 seconds
    write_bytes_per_second = (
        desires.query_pattern.estimated_write_per_second.mid
        * desires.query_pattern.estimated_mean_write_size_bytes.mid
    )

    compactions_per_hour = 2
    hour_in_seconds = 60 * 60

    write_buffer_gib = (
        (write_bytes_per_second * hour_in_seconds)
        / (flushes_before_compaction**compactions_per_hour)
    ) / (1 << 30)

    return float(write_buffer_gib) / zones_per_region


def _get_cores_from_desires(desires: CapacityDesires, instance: Instance) -> int:
    cpu_buffer = buffer_for_components(
        buffers=desires.buffers, components=[BACKGROUND_BUFFER]
    )

    # We have no existing utilization to go from
    reference_shape = desires.reference_shape
    # Keep half of the cores free for background work (compaction, backup, repair).
    needed_cores = math.ceil(sqrt_staffed_cores(desires) * cpu_buffer.ratio)

    needed_cores = normalize_cores(
        core_count=needed_cores,
        target_shape=instance,
        reference_shape=reference_shape,
    )
    return needed_cores


def _get_disk_from_desires(desires: CapacityDesires, copies_per_region: int) -> int:
    disk_buffer = buffer_for_components(
        buffers=desires.buffers, components=[BufferComponent.disk]
    )
    # Do not add disk buffers now as memory calculation is done on the disk usage
    return math.ceil(
        (1.0 / desires.data_shape.estimated_compression_ratio.mid)
        * desires.data_shape.estimated_state_size_gib.mid
        * copies_per_region
        * disk_buffer.ratio
    )


def _get_min_count(
    tier: int,
    required_cluster_size: Optional[int],
    needed_disk_gib: float,
    disk_per_node_gib: float,
    cluster_size_lambda: Callable[[int], int],
) -> int:
    """
    Compute the minimum number of nodes required for a zone.

    This function is used to prevent the planner from allocating clusters that
    would exceed the max data per node or under the required cluster size for
    a tier or existing cluster
    """

    # Cassandra clusters should aim to be at least 2 nodes per zone to start
    # out with for tier 0 or tier 1. This gives us more room to "up-color"]
    # clusters.
    min_nodes_for_tier = 2 if tier in CRITICAL_TIERS else 0

    # Prevent allocating clusters that exceed the max data per node.
    min_nodes_for_disk = math.ceil(needed_disk_gib / disk_per_node_gib)

    # Take the max of the following in order to avoid:
    # (1) if `required_cluster_size` < `min_nodes_for_disk`, don't let the planner
    #     pick a shape that would exceed the max data per node
    #
    #     For example, if we need 4TiB of disk, and the max data per node is 1TiB,
    #     Regardless of the `required_cluster_size`, we cannot allocate less than 4
    #     nodes because that would exceed the max data per node.
    #
    # (2) if `required_cluster_size` > `min_nodes_for_disk`, don't let the
    #     node density requirement affect the min count because the required
    #     cluster size already meets the node density requirement.
    #
    #     For example, if we need 4TiB of disk, and the max data per node is 1TiB,
    #     and the upstream requires >= 8 nodes, we can allocate 8 nodes because
    #     each node would only have 500GB of data.
    min_count = max(
        min_nodes_for_tier,
        required_cluster_size or 0,
        min_nodes_for_disk,
    )
    # Ensure that the min count is an increment of the cluster size constraint (doubling)
    return cluster_size_lambda(min_count)


def _zonal_requirement_for_new_cluster(
    desires: CapacityDesires,
    instance: Instance,
    copies_per_region: int,
    zones_per_region: int,
) -> CapacityRequirement:
    needed_cores = _get_cores_from_desires(desires, instance)
    needed_disk = _get_disk_from_desires(desires, copies_per_region)

    # Keep some of the bandwidth available for backup and repair streaming
    network_buffer = buffer_for_components(
        buffers=desires.buffers, components=[BACKGROUND_BUFFER]
    )
    needed_network_mbps = simple_network_mbps(desires) * network_buffer.ratio

    # Now convert to per zone
    needed_cores = max(1, needed_cores // zones_per_region)
    needed_disk = max(1, needed_disk // zones_per_region)

    return CapacityRequirement(
        requirement_type="zonal-capacity",
        cpu_cores=certain_int(needed_cores),
        disk_gib=certain_float(needed_disk),
        network_mbps=certain_float(needed_network_mbps),
    )


# pylint: disable=too-many-locals
# pylint: disable=too-many-positional-arguments
def _estimate_cassandra_requirement(
    instance: Instance,
    desires: CapacityDesires,
    working_set: float,
    reads_per_second: float,
    max_rps_to_disk: int,
    zones_per_region: int = 3,
    copies_per_region: int = 3,
) -> CapacityRequirement:
    """Estimate the capacity required for one zone given a regional desire

    The input desires should be the **regional** desire, and this function will
    return the zonal capacity requirement
    """
    disk_buffer = buffer_for_components(
        buffers=desires.buffers, components=[BufferComponent.disk]
    )
    memory_preserve = False
    reference_shape = desires.reference_shape
    current_capacity = _get_current_capacity(desires)

    # If the cluster is already provisioned
    if current_capacity and desires.current_clusters is not None:
        capacity_requirement = zonal_requirements_from_current(
            desires.current_clusters,
            desires.buffers,
            instance,
            reference_shape,
        )
        disk_derived_buffer = DerivedBuffers.for_components(
            desires.buffers.derived, [BufferComponent.disk]
        )
        disk_used_gib = (
            current_capacity.disk_utilization_gib.mid
            * current_capacity.cluster_instance_count.mid
            * disk_derived_buffer.scale
        )
        memory_preserve = DerivedBuffers.for_components(
            desires.buffers.derived, [BufferComponent.memory]
        ).preserve
    else:
        # If the cluster is not yet provisioned
        capacity_requirement = _zonal_requirement_for_new_cluster(
            desires, instance, copies_per_region, zones_per_region
        )
        disk_used_gib = capacity_requirement.disk_gib.mid / disk_buffer.ratio

    needed_cores = math.ceil(capacity_requirement.cpu_cores.mid)
    needed_disk = capacity_requirement.disk_gib.mid
    needed_network_mbps = capacity_requirement.network_mbps.mid

    # it can be 0 for cases where disk utilization is not passed as a part of current cluster capacity
    if needed_disk == 0:
        needed_disk = max(
            1, _get_disk_from_desires(desires, copies_per_region) // zones_per_region
        )
        # We want to compute memory based on the data size.
        disk_used_gib = needed_disk / disk_buffer.ratio

    # Rough estimate of how many instances we would need just for the CPU
    # Note that this is a lower bound, we might end up with more.
    estimated_cores_per_region = math.ceil(
        (needed_cores * zones_per_region) / instance.cpu
    )
    # Generally speaking we want fewer than some number of reads per second
    # hitting disk per instance. If we don't have many reads we don't need to
    # hold much data in memory.
    instance_rps = max(1, reads_per_second // estimated_cores_per_region)
    disk_rps = instance_rps * _cass_io_per_read(
        max(1, (disk_used_gib * zones_per_region) // estimated_cores_per_region)
    )
    rps_working_set = min(1.0, disk_rps / max_rps_to_disk)

    # Cassandra can defer writes either by buffering in memory or by
    # waiting longer before recompacting (the min-threshold on the
    # L0 compactions or STCS compactions)
    min_threshold = 4
    write_buffer_gib = _write_buffer_gib_zone(
        desires=desires,
        zones_per_region=zones_per_region,
        flushes_before_compaction=min_threshold,
    )

    while write_buffer_gib > 12 and min_threshold < 16:
        min_threshold *= 2
        write_buffer_gib = _write_buffer_gib_zone(
            desires=desires,
            zones_per_region=zones_per_region,
            flushes_before_compaction=min_threshold,
        )

    if current_capacity and current_capacity.cluster_instance and memory_preserve:
        # remove base memory and heap from per node ram and then
        # multiply by number of nodes in a zone to compute the zonal requirement.
        reserve_memory = _get_base_memory(desires) + _cass_heap(
            current_capacity.cluster_instance.ram_gib
        )
        needed_memory = (
            current_capacity.cluster_instance.ram_gib - reserve_memory
        ) * current_capacity.cluster_instance_count.mid
        write_buffer_gib = 0
    else:
        # If disk RPS will be smaller than our target because there are no
        # reads, we don't need to hold as much data in memory.
        # For c*, we can skip memory buffer and can just keep using the heap and write buffer calc
        # Eventually we'll want to phrase those heap, read cache, and write cache as buffers
        needed_memory = (
            min(working_set, rps_working_set) * disk_used_gib * zones_per_region
        )
        # Now convert to per zone
        needed_memory = max(1, int(needed_memory // zones_per_region))

    logger.debug(
        "Need (cpu, mem, disk, working) = (%s, %s, %s, %f)",
        needed_cores,
        needed_memory,
        needed_disk,
        working_set,
    )

    return CapacityRequirement(
        requirement_type="cassandra-zonal",
        reference_shape=reference_shape,
        cpu_cores=certain_int(needed_cores),
        mem_gib=certain_float(needed_memory),
        disk_gib=certain_float(needed_disk),
        network_mbps=certain_float(needed_network_mbps),
        context={
            "working_set": min(working_set, rps_working_set),
            "rps_working_set": rps_working_set,
            "disk_slo_working_set": working_set,
            "replication_factor": copies_per_region,
            "compression_ratio": round(
                1.0 / desires.data_shape.estimated_compression_ratio.mid, 2
            ),
            "read_per_second": reads_per_second,
            "write_buffer_gib": write_buffer_gib,
            "min_threshold": min_threshold,
        },
    )


def _get_current_cluster_size(desires: CapacityDesires) -> int:
    current_capacity = _get_current_capacity(desires)
    if current_capacity is None:
        return 0
    return math.ceil(current_capacity.cluster_instance_count.mid)


def _get_current_capacity(desires: CapacityDesires) -> Optional[CurrentClusterCapacity]:
    current_capacity = (
        None
        if desires.current_clusters is None
        else (
            desires.current_clusters.zonal[0]
            if len(desires.current_clusters.zonal)
            else desires.current_clusters.regional[0]
        )
    )
    return current_capacity


def _upsert_params(cluster: Any, params: Dict[str, Any]) -> None:
    if cluster.cluster_params:
        cluster.cluster_params.update(params)
    else:
        cluster.cluster_params = params


def _get_cluster_size_lambda(
    current_cluster_size: int,
    required_cluster_size: Optional[int],
) -> Callable[[int], int]:
    if required_cluster_size:
        return lambda x: next_doubling(x, base=required_cluster_size)
    elif current_cluster_size and not is_power_of_2(current_cluster_size):
        return lambda x: next_doubling(x, base=current_cluster_size)
    else:  # New provisionings
        return next_power_of_2


# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
# flake8: noqa: C901
def _estimate_cassandra_cluster_zonal(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    drive: Drive,
    context: RegionContext,
    desires: CapacityDesires,
    zones_per_region: int = 3,
    copies_per_region: int = 3,
    require_local_disks: bool = False,
    require_attached_disks: bool = False,
    required_cluster_size: Optional[int] = None,
    max_rps_to_disk: int = 500,
    max_local_data_per_node_gib: int = 1280,
    max_attached_data_per_node_gib: int = 2048,
    max_regional_size: int = 192,
    max_write_buffer_percent: float = 0.25,
    max_table_buffer_percent: float = 0.11,
) -> Optional[CapacityPlan]:
    # Netflix Cassandra doesn't like to deploy on really small instances
    if instance.cpu < 2 or instance.ram_gib < 14:
        return None

    # if we're not allowed to use gp2, skip EBS only types
    if instance.drive is None and require_local_disks:
        return None

    # if we're not allowed to use local disks, skip ephems
    if instance.drive is not None and require_attached_disks:
        return None

    # Cassandra only deploys on gp2 and gp3 drives right now
    if drive.name not in ("gp2", "gp3"):
        return None

    rps = desires.query_pattern.estimated_read_per_second.mid // zones_per_region
    write_per_sec = (
        desires.query_pattern.estimated_write_per_second.mid // zones_per_region
    )
    write_bytes_per_sec = round(
        write_per_sec * desires.query_pattern.estimated_mean_write_size_bytes.mid
    )
    read_bytes_per_sec = rps * desires.query_pattern.estimated_mean_read_size_bytes.mid
    # Write IO will be 1 to commitlog + 2 writes (plus 2 reads) in the first
    # hour during compaction.
    # Writes are sequential
    # Reads are random
    write_io_per_sec = (1 + 4) * max(
        1, write_bytes_per_sec // (drive.seq_io_size_kib * 1024)
    )
    read_io_per_sec = max(rps, read_bytes_per_sec // (drive.rand_io_size_kib * 1024))

    # Based on the disk latency and the read latency SLOs we adjust our
    # working set to keep more or less data in RAM. Faster drives need
    # less fronting RAM.
    ws_drive = instance.drive or drive
    working_set = working_set_from_drive_and_slo(
        drive_read_latency_dist=dist_for_interval(ws_drive.read_io_latency_ms),
        read_slo_latency_dist=dist_for_interval(
            desires.query_pattern.read_latency_slo_ms
        ),
        estimated_working_set=desires.data_shape.estimated_working_set_percent,
        # This is about right for a database, a cache probably would want
        # to increase this even more.
        target_percentile=0.95,
    ).mid

    requirement = _estimate_cassandra_requirement(
        instance=instance,
        desires=desires,
        working_set=working_set,
        reads_per_second=rps,
        max_rps_to_disk=max_rps_to_disk,
        zones_per_region=zones_per_region,
        copies_per_region=copies_per_region,
    )

    # Adjust the min count to adjust to prevent too much data on a single
    needed_disk_gib = int(requirement.disk_gib.mid)
    disk_buffer_ratio = buffer_for_components(
        buffers=desires.buffers, components=[BufferComponent.disk]
    ).ratio
    disk_per_node_gib = get_effective_disk_per_node_gib(
        instance,
        drive,
        disk_buffer_ratio,
        max_local_data_per_node_gib=max_local_data_per_node_gib,
        max_attached_data_per_node_gib=max_attached_data_per_node_gib,
    )

    current_cluster_size = _get_current_cluster_size(desires)
    cluster_size_lambda = _get_cluster_size_lambda(
        current_cluster_size, required_cluster_size
    )
    min_count = _get_min_count(
        tier=desires.service_tier,
        required_cluster_size=required_cluster_size,
        needed_disk_gib=needed_disk_gib,
        disk_per_node_gib=disk_per_node_gib,
        cluster_size_lambda=cluster_size_lambda,
    )

    base_mem = _get_base_memory(desires)

    heap_fn = _cass_heap_for_write_buffer(
        instance=instance,
        max_zonal_size=max_regional_size // zones_per_region,
        write_buffer_gib=requirement.context["write_buffer_gib"],
        buffer_percent=(max_write_buffer_percent * max_table_buffer_percent),
    )

    cluster = compute_stateful_zone(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=needed_disk_gib,
        needed_memory_gib=int(requirement.mem_gib.mid),
        needed_network_mbps=requirement.network_mbps.mid,
        # Take into account the reads per read
        # from the per node dataset using leveled compaction
        required_disk_ios=lambda size, count: (
            _cass_io_per_read(size) * math.ceil(read_io_per_sec / count),
            write_io_per_sec / count,
        ),
        # C* clusters provision in powers of 2 because doubling
        cluster_size=cluster_size_lambda,
        min_count=min_count,
        # TODO: Take reserve memory calculation into account during buffer calculation
        # C* heap usage takes away from OS page cache memory
        reserve_memory=lambda x: base_mem + heap_fn(x),
        # C* heap buffers the writes at roughly a rate of
        # memtable_cleanup_threshold * memtable_size. At Netflix this
        # is 0.11 * 25 * heap
        write_buffer=lambda x: heap_fn(x) * max_write_buffer_percent * 0.25,
        required_write_buffer_gib=float(requirement.context["write_buffer_gib"]),
    )

    # Communicate to the actual provision that if we want reduced RF
    params = {
        "cassandra.keyspace.rf": copies_per_region,
        # In order to handle high write loads we have to shift memory
        # to heap memory, communicate with C* about this
        "cassandra.heap.gib": heap_fn(instance.ram_gib),
        "cassandra.heap.write.percent": max_write_buffer_percent,
        "cassandra.heap.table.percent": max_table_buffer_percent,
        "cassandra.compaction.min_threshold": requirement.context["min_threshold"],
    }
    _upsert_params(cluster, params)

    # Sometimes we don't want modify cluster topology, so only allow
    # topologies that match the desired zone size
    if required_cluster_size is not None and cluster.count != required_cluster_size:
        return None

    # Cassandra clusters generally should try to stay under some total number
    # of nodes. Orgs do this for all kinds of reasons such as
    #   * Security group limits. Since you must have < 500 rules if you're
    #       ingressing public ips)
    #   * Maintenance. If your restart script does one node at a time you want
    #       smaller clusters so your restarts don't take months.
    #   * Schema propagation. Since C* must gossip out changes to schema the
    #       duration of this can increase a lot with > 500 node clusters.
    if cluster.count > (max_regional_size // zones_per_region):
        return None

    # Durable Cassandra clusters backup to S3
    # TODO use the write rate and estimated write size to estimate churn
    # over the retention period.
    cap_services = []
    if desires.data_shape.durability_slo_order.mid >= 1000:
        blob = context.services.get("blob.standard", None)
        if blob:
            cap_services = [
                ServiceCapacity(
                    service_type=f"cassandra.backup.{blob.name}",
                    annual_cost=blob.annual_cost_gib(requirement.disk_gib.mid),
                    service_params={
                        "nines_required": (
                            1 - 1.0 / desires.data_shape.durability_slo_order.mid
                        )
                    },
                )
            ]

    network_costs = network_services("cassandra", context, desires, copies_per_region)
    if network_costs:
        cap_services.extend(network_costs)

    # Account for the clusters, backup, and network costs
    cassandra_costs = {
        "cassandra.zonal-clusters": zones_per_region * cluster.annual_cost,
    }
    for s in cap_services:
        cassandra_costs[f"{s.service_type}"] = s.annual_cost

    cluster.cluster_type = "cassandra"
    clusters = Clusters(
        annual_costs=cassandra_costs,
        zonal=[cluster] * zones_per_region,
        regional=[],
        services=cap_services,
    )

    return CapacityPlan(
        requirements=Requirements(zonal=[requirement] * zones_per_region),
        candidate_clusters=clusters,
    )


# C* LCS has 160 MiB sstables by default and 10 sstables per level
def _cass_io_per_read(node_size_gib: float, sstable_size_mb: int = 160) -> int:
    gb = node_size_gib * 1024
    sstables = max(1, gb // sstable_size_mb)
    # 10 sstables per level, plus 1 for L0 (avg)
    levels = 1 + int(math.ceil(math.log(sstables, 10)))
    # One disk IO per data read and one per index read (assume we miss
    # the key cache)
    return 2 * levels


def _get_base_memory(desires: CapacityDesires) -> float:
    return (
        desires.data_shape.reserved_instance_app_mem_gib
        + desires.data_shape.reserved_instance_system_mem_gib
    )


def _cass_heap_for_write_buffer(
    instance: Instance,
    write_buffer_gib: float,
    max_zonal_size: int,
    buffer_percent: float,
) -> Callable[[float], float]:
    # If there is no way we can get enough heap with the max zonal size, try
    # letting max heap grow to 31 GiB per node to get more write buffer
    if write_buffer_gib > (
        max_zonal_size * _cass_heap(instance.ram_gib) * buffer_percent
    ):
        return lambda x: _cass_heap(x, max_heap_gib=30)
    else:
        return _cass_heap


# C* follows the following formula for calculating heap
def _cass_heap(node_memory_gib: float, max_heap_gib: float = 30) -> float:
    # OSS Cassandra does this
    # max(min(node_memory_gib // 2, 4), min(node_memory_gib // 4, max_heap_gib))

    # Netflix Cassandra does this
    return min(max(4, node_memory_gib // 2), max_heap_gib)


def _target_rf(desires: CapacityDesires, user_copies: Optional[int]) -> int:
    if user_copies is not None:
        assert user_copies > 1
        return user_copies

    # Due to the relaxed durability and consistency requirements we can
    # run with RF=2
    consistency = desires.query_pattern.access_consistency.same_region
    if (
        desires.data_shape.durability_slo_order.mid < 1000
        and consistency is not None
        and consistency.target_consistency != AccessConsistency.read_your_writes
    ):
        return 2
    return 3


class NflxCassandraArguments(BaseModel):
    """Configuration arguments for the Netflix Cassandra capacity model.

    This model centralizes all tunable parameters with their defaults.
    Use `from_extra_model_arguments()` to parse a dict into a validated instance.
    """

    copies_per_region: Optional[int] = Field(
        default=None,
        description="How many copies of the data will exist e.g. RF=3. If unsupplied"
        " this will be deduced from durability and consistency desires",
    )
    require_local_disks: bool = Field(
        default=False,
        description="If local (ephemeral) drives are required",
    )
    require_attached_disks: bool = Field(
        default=False,
        description="If attached (ebs) drives are required",
    )
    required_cluster_size: Optional[int] = Field(
        default=None,
        description="Require zonal clusters to be this size (force vertical scaling)",
    )
    max_rps_to_disk: int = Field(
        default=500,
        description="How many disk IOs should be allowed to hit disk per instance",
    )
    max_regional_size: int = Field(
        default=192,
        description="What is the maximum size of a cluster in this region",
    )
    max_local_data_per_node_gib: int = Field(
        default=1280,
        description="Maximum data per node for local disk instances (GiB)",
    )
    max_attached_data_per_node_gib: int = Field(
        default=2048,
        description="Maximum data per node for attached disk instances (GiB)",
    )
    max_write_buffer_percent: float = Field(
        default=0.25,
        description="The amount of heap memory that can be used to buffer writes. "
        "Note that if there are more than 100k writes this will "
        "automatically adjust to 0.5",
    )
    max_table_buffer_percent: float = Field(
        default=0.11,
        description="How much of heap memory can be used for a single table. "
        "Note that if there are more than 100k writes this will "
        "automatically adjust to 0.2",
    )

    @classmethod
    def from_extra_model_arguments(
        cls, extra_model_arguments: Dict[str, Any]
    ) -> "NflxCassandraArguments":
        """Parse extra_model_arguments dict into a validated NflxCassandraArguments.

        This centralizes default values - any field not in extra_model_arguments
        will use the default defined in this model.

        Handles legacy field name mappings:
        - max_local_disk_gib -> max_local_data_per_node_gib (if not explicitly set)
        """
        # Handle legacy field name: max_local_disk_gib -> max_local_data_per_node_gib
        args = dict(extra_model_arguments)
        if "max_local_data_per_node_gib" not in args and "max_local_disk_gib" in args:
            args["max_local_data_per_node_gib"] = args["max_local_disk_gib"]

        # Pydantic will use defaults for any missing fields
        return cls.model_validate(args)


class NflxCassandraCapacityModel(CapacityModel):
    def __init__(self) -> None:
        pass

    @staticmethod
    def get_required_cluster_size(
        tier: int, extra_model_arguments: Dict[str, Any]
    ) -> Optional[int]:
        required_cluster_size: Optional[int] = (
            math.ceil(extra_model_arguments["required_cluster_size"])
            if "required_cluster_size" in extra_model_arguments
            else None
        )

        if tier not in CRITICAL_TIERS or required_cluster_size is None:
            return required_cluster_size

        # If the upstream explicitly set a cluster size, make sure it is
        # at least CRITICAL_TIER_MIN_CLUSTER_SIZE. We cannot do a max
        # of the two because the horizontal scaling is disabled
        if required_cluster_size < CRITICAL_TIER_MIN_CLUSTER_SIZE:
            raise ValueError(
                f"Required cluster size must be at least "
                f"{CRITICAL_TIER_MIN_CLUSTER_SIZE=} when "
                f"service tier({tier}) is a "
                f"critical tier({CRITICAL_TIERS}). "
                f"If it is an existing cluster, horizontally "
                f"scale the cluster to be >= "
                f"{CRITICAL_TIER_MIN_CLUSTER_SIZE}"
            )

        return required_cluster_size

    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        # Parse extra_model_arguments into a validated model with centralized defaults
        args = NflxCassandraArguments.from_extra_model_arguments(extra_model_arguments)

        # Use durability and consistency to compute RF if not explicitly set
        copies_per_region = _target_rf(desires, args.copies_per_region)

        # Validate required_cluster_size for critical tiers
        required_cluster_size: Optional[int] = (
            NflxCassandraCapacityModel.get_required_cluster_size(
                desires.service_tier, extra_model_arguments
            )
        )

        # Apply caps to buffer percentages
        max_write_buffer_percent = min(0.5, args.max_write_buffer_percent)
        max_table_buffer_percent = min(0.5, args.max_table_buffer_percent)

        # Adjust heap defaults for high write clusters
        if (
            desires.query_pattern.estimated_write_per_second.mid >= 100_000
            and desires.data_shape.estimated_state_size_gib.mid >= 100
        ):
            max_write_buffer_percent = max(0.5, max_write_buffer_percent)
            max_table_buffer_percent = max(0.2, max_table_buffer_percent)

        return _estimate_cassandra_cluster_zonal(
            instance=instance,
            drive=drive,
            context=context,
            desires=desires,
            zones_per_region=context.zones_in_region,
            copies_per_region=copies_per_region,
            require_local_disks=args.require_local_disks,
            require_attached_disks=args.require_attached_disks,
            required_cluster_size=required_cluster_size,
            max_rps_to_disk=args.max_rps_to_disk,
            max_regional_size=args.max_regional_size,
            max_local_data_per_node_gib=args.max_local_data_per_node_gib,
            max_write_buffer_percent=max_write_buffer_percent,
            max_table_buffer_percent=max_table_buffer_percent,
        )

    @staticmethod
    def description() -> str:
        return "Netflix Streaming Cassandra Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxCassandraArguments.model_json_schema()

    @staticmethod
    def default_buffers() -> Buffers:
        return Buffers(
            default=Buffer(ratio=1.5),
            desired={
                "compute": Buffer(ratio=1.5, components=[BufferComponent.compute]),
                "storage": Buffer(ratio=4.0, components=[BufferComponent.storage]),
                # Cassandra reserves headroom in both cpu and network for background
                # work and tasks
                "background": Buffer(
                    ratio=2.0,
                    components=[
                        BufferComponent.cpu,
                        BufferComponent.network,
                        BACKGROUND_BUFFER,
                    ],
                ),
            },
        )

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        acceptable_consistency = {
            None,
            AccessConsistency.best_effort,
            AccessConsistency.eventual,
            AccessConsistency.read_your_writes,
            AccessConsistency.never,
        }
        for key, value in user_desires.query_pattern.access_consistency:
            if value.target_consistency not in acceptable_consistency:
                raise ValueError(
                    f"Cassandra can only provide {acceptable_consistency} access."
                    f"User asked for {key}={value}"
                )

        # Lower RF = less write compute
        rf = _target_rf(
            user_desires, extra_model_arguments.get("copies_per_region", None)
        )
        if rf < 3:
            rf_write_latency = Interval(low=0.2, mid=0.6, high=2, confidence=0.98)
        else:
            rf_write_latency = Interval(low=0.4, mid=1, high=2, confidence=0.98)

        # By supplying these buffers we can deconstruct observed utilization into
        # load versus buffer.
        buffers = NflxCassandraCapacityModel.default_buffers()
        if user_desires.query_pattern.access_pattern == AccessPattern.latency:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.latency,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=64, mid=256, high=1024, confidence=0.95
                    ),
                    # Cassandra point queries usualy take just around 2ms
                    # of on CPU time for reads and 1ms for writes
                    estimated_mean_read_latency_ms=Interval(
                        low=0.4, mid=2, high=5, confidence=0.98
                    ),
                    estimated_mean_write_latency_ms=rf_write_latency,
                    # Assume point queries, "Single digit milliseconds SLO"
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=2,
                        high=5,
                        confidence=0.98,
                    ),
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.2,
                        maximum_value=10,
                        low=0.4,
                        mid=1,
                        high=4,
                        confidence=0.98,
                    ),
                ),
                # Most latency sensitive cassandra clusters are in the
                # < 1TiB range
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=10, mid=100, high=1000, confidence=0.98
                    ),
                    # Cassandra compresses with LZ4 by default
                    estimated_compression_ratio=Interval(
                        minimum_value=1.1,
                        maximum_value=8,
                        low=2,
                        mid=3,
                        high=5,
                        confidence=0.98,
                    ),
                    # We dynamically allocate the C* JVM memory in the plan
                    # but account for the Priam sidecar here
                    reserved_instance_app_mem_gib=4,
                ),
                buffers=buffers,
            )
        else:
            return CapacityDesires(
                query_pattern=QueryPattern(
                    access_pattern=AccessPattern.throughput,
                    access_consistency=GlobalConsistency(
                        same_region=Consistency(
                            target_consistency=AccessConsistency.read_your_writes,
                        ),
                        cross_region=Consistency(
                            target_consistency=AccessConsistency.eventual,
                        ),
                    ),
                    estimated_mean_read_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    estimated_mean_write_size_bytes=Interval(
                        low=128, mid=1024, high=65536, confidence=0.95
                    ),
                    # Cassandra scan queries usually take longer
                    estimated_mean_read_latency_ms=Interval(
                        low=0.2, mid=5, high=20, confidence=0.98
                    ),
                    # Usually throughput clusters are running RF=2
                    # Maybe revise this?
                    estimated_mean_write_latency_ms=Interval(
                        low=0.2, mid=0.6, high=2, confidence=0.98
                    ),
                    # Assume they're scanning -> slow reads
                    read_latency_slo_ms=FixedInterval(
                        minimum_value=1,
                        maximum_value=100,
                        low=2,
                        mid=8,
                        high=90,
                        confidence=0.98,
                    ),
                    # Assume they're doing BATCH writes
                    write_latency_slo_ms=FixedInterval(
                        minimum_value=0.5,
                        maximum_value=20,
                        low=1,
                        mid=2,
                        high=8,
                        confidence=0.98,
                    ),
                ),
                data_shape=DataShape(
                    estimated_state_size_gib=Interval(
                        low=100, mid=1000, high=4000, confidence=0.98
                    ),
                    # Cassandra compresses with LZ4 by default
                    estimated_compression_ratio=Interval(
                        low=2, mid=3, high=5, confidence=0.98
                    ),
                    # We dynamically allocate the C* JVM memory in the plan
                    # but account for the Priam sidecar here
                    reserved_instance_app_mem_gib=4,
                ),
                buffers=buffers,
            )


nflx_cassandra_capacity_model = NflxCassandraCapacityModel()



from service_capacity_modeling.models.headroom_strategy import (
    QueuingBasedHeadroomStrategy,
)


def test_calculate_headroom_using_default_strategy():
    effective_cpu_small_instance = 4
    effective_cpu_large_instance = 16
    headroom_strategy = QueuingBasedHeadroomStrategy()
    reserved_headroom_small_instance = headroom_strategy.calculate_reserved_headroom(
        effective_cpu_small_instance
    )
    reserved_headroom_large_instance = headroom_strategy.calculate_reserved_headroom(
        effective_cpu_large_instance
    )

    # The headroom should be inversely proportional to effective CPU
    assert reserved_headroom_large_instance < reserved_headroom_small_instance



import logging
import math
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from pydantic import BaseModel
from pydantic import Field

from service_capacity_modeling.enum_utils import StrEnum
from service_capacity_modeling.interface import AccessConsistency
from service_capacity_modeling.interface import AccessPattern
from service_capacity_modeling.interface import Buffer
from service_capacity_modeling.interface import BufferComponent
from service_capacity_modeling.interface import Buffers
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import CapacityPlan
from service_capacity_modeling.interface import CapacityRequirement
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import Clusters
from service_capacity_modeling.interface import Consistency
from service_capacity_modeling.interface import CurrentZoneClusterCapacity
from service_capacity_modeling.interface import DataShape
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import GIB_IN_BYTES
from service_capacity_modeling.interface import GlobalConsistency
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import Interval
from service_capacity_modeling.interface import MEGABIT_IN_BYTES
from service_capacity_modeling.interface import MIB_IN_BYTES
from service_capacity_modeling.interface import QueryPattern
from service_capacity_modeling.interface import RegionContext
from service_capacity_modeling.interface import Requirements
from service_capacity_modeling.models import CapacityModel
from service_capacity_modeling.models.common import buffer_for_components
from service_capacity_modeling.models.common import compute_stateful_zone
from service_capacity_modeling.models.common import get_effective_disk_per_node_gib
from service_capacity_modeling.models.common import normalize_cores
from service_capacity_modeling.models.common import sqrt_staffed_cores
from service_capacity_modeling.models.common import zonal_requirements_from_current
from service_capacity_modeling.models.org.netflix.iso_date_math import iso_to_seconds

logger = logging.getLogger(__name__)


class ClusterType(StrEnum):
    strong = "strong"
    ha = "high-availability"


def _get_current_zonal_cluster(
    desires: CapacityDesires,
) -> Optional[CurrentZoneClusterCapacity]:
    return (
        None
        if desires.current_clusters is None
        else (
            desires.current_clusters.zonal[0]
            if len(desires.current_clusters.zonal)
            else None
        )
    )


def _is_same_instance_family(
    cluster: Optional[CurrentZoneClusterCapacity], target_family: str
) -> bool:
    """Check if cluster has a different instance family than the target."""
    if not cluster or not cluster.cluster_instance:
        return False
    return cluster.cluster_instance.family == target_family


def _estimate_kafka_requirement(  # pylint: disable=too-many-positional-arguments
    instance: Instance,
    desires: CapacityDesires,
    copies_per_region: int,
    hot_retention_seconds: float,
    zones_per_region: int = 3,
) -> Tuple[CapacityRequirement, Tuple[str, ...]]:
    """Estimate the capacity required for one zone given a regional desire

    The input desires should be the **regional** desire, and this function will
    return the zonal capacity requirement
    """
    normalized_to_mib = desires.model_copy(deep=True)
    read_mib_per_second = (
        normalized_to_mib.query_pattern.estimated_mean_read_size_bytes.mid
        / MIB_IN_BYTES
    )
    write_mib_per_second = (
        normalized_to_mib.query_pattern.estimated_mean_write_size_bytes.mid
        / MIB_IN_BYTES
    )

    # 1 concurrent reader, cpu time is per MiB read
    # 0.5 MiB / second
    rps = normalized_to_mib.query_pattern.estimated_read_per_second
    wps = normalized_to_mib.query_pattern.estimated_write_per_second
    normalized_to_mib.query_pattern.estimated_read_per_second = rps.scale(
        read_mib_per_second
    )
    normalized_to_mib.query_pattern.estimated_write_per_second = wps.scale(
        write_mib_per_second
    )
    # use the current cluster capacity if available
    current_zonal_capacity = _get_current_zonal_cluster(desires)

    bw_in = (
        (write_mib_per_second * MIB_IN_BYTES) * copies_per_region
    ) / MEGABIT_IN_BYTES
    bw_out = (
        (read_mib_per_second * MIB_IN_BYTES)
        + ((write_mib_per_second * MIB_IN_BYTES) * (copies_per_region - 1))
    ) / MEGABIT_IN_BYTES
    if (
        current_zonal_capacity
        and current_zonal_capacity.cluster_instance
        and desires.current_clusters is not None
    ):
        # zonal_requirements_from_current uses the midpoint utilization of the
        # current cluster. For Kafka, we want to use the high value instead
        # for cpu, disk, network, etc.
        normalize_midpoint_desires = desires.model_copy(deep=True)
        # This is checked in the if statement above. Assert here for mypy linter purpose
        assert normalize_midpoint_desires.current_clusters is not None
        curr_disk = desires.current_clusters.zonal[0].disk_utilization_gib
        curr_cpu = desires.current_clusters.zonal[0].cpu_utilization
        curr_network = desires.current_clusters.zonal[0].network_utilization_mbps
        normalize_midpoint_desires.current_clusters.zonal[
            0
        ].disk_utilization_gib = Interval(
            low=curr_disk.high, mid=curr_disk.high, high=curr_disk.high
        )
        normalize_midpoint_desires.current_clusters.zonal[0].cpu_utilization = Interval(
            low=curr_cpu.high, mid=curr_cpu.high, high=curr_cpu.high
        )
        normalize_midpoint_desires.current_clusters.zonal[
            0
        ].network_utilization_mbps = Interval(
            low=curr_network.high, mid=curr_network.high, high=curr_network.high
        )
        capacity_requirement = zonal_requirements_from_current(
            current_cluster=normalize_midpoint_desires.current_clusters,
            buffers=desires.buffers,
            instance=instance,
            reference_shape=current_zonal_capacity.cluster_instance,
        )
        needed_cores = int(capacity_requirement.cpu_cores.mid)
        needed_disk = int(capacity_requirement.disk_gib.mid)
        needed_network_mbps = int(capacity_requirement.network_mbps.mid)
        logger.debug(
            "kafka normalized needed cores: %s", capacity_requirement.cpu_cores
        )
    else:
        # We have no existing utilization to go from
        needed_cores = (
            normalize_cores(
                core_count=sqrt_staffed_cores(normalized_to_mib),
                target_shape=instance,
                reference_shape=desires.reference_shape,
            )
            // zones_per_region
        )

        # (Nick): Keep 40% of available bandwidth for node recovery
        # (Joey): For kafka BW = BW_write + BW_reads
        #   let X = input write BW
        #   BW_in = X * RF
        #   BW_out = X * (consumers) + X * (RF - 1)
        #   BW = (in + out) because duplex then 40% headroom.
        needed_network_mbps = int((max(bw_in, bw_out) * 1.40) // zones_per_region)

        # NOTE: data_shape is region, we need to convert it to zonal
        # If we don't have an existing cluster, the estimated state size should be
        # at most 40% of the total cluster's available disk.
        # i.e. needed_disk = state_size * 2.5
        needed_disk = math.ceil(
            (desires.data_shape.estimated_state_size_gib.mid // zones_per_region) * 2.5,
        )

    # Keep the last N seconds hot in cache
    needed_memory = (
        (write_mib_per_second * hot_retention_seconds) // 1024
    ) // zones_per_region

    logger.debug(
        "Need (instance, cpu, mem, disk) = (%s, %s, %s, %s)",
        instance.name,
        needed_cores,
        needed_memory,
        needed_disk,
    )

    return (
        CapacityRequirement(
            requirement_type="kafka-zonal",
            reference_shape=desires.reference_shape,
            cpu_cores=certain_int(needed_cores),
            mem_gib=certain_float(needed_memory),
            disk_gib=certain_float(needed_disk),
            network_mbps=certain_float(needed_network_mbps),
            context={
                "bw_in_mbps": bw_in,
                "bw_out_mbps": bw_out,
                "hot_retention_seconds": hot_retention_seconds,
                "replication_factor": copies_per_region,
            },
        ),
        ("spend", "disk", "mem"),
    )


def _upsert_params(cluster: Any, params: Dict[str, Any]) -> None:
    if cluster.cluster_params:
        cluster.cluster_params.update(params)
    else:
        cluster.cluster_params = params


def _kafka_read_io(
    rps: float, io_size_kib: float, size_gib: float, recovery_seconds: int
) -> float:
    # Get enough disk read IO capacity for some reads
    # In practice we have cache reducing this by 99% or more
    read_ios = rps * 0.05
    # Recover the node in 60 minutes, to do that we need
    # Estimate that we are using ~ 50% of the disk prior to a recovery
    size_kib = size_gib * 0.5 * (1024 * 1024)
    recovery_ios = max(1, size_kib / io_size_kib) / recovery_seconds
    # Leave 50% headroom for read IOs since generally we will hit cache
    return float((read_ios + int(round(recovery_ios))) * 1.5)


# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-positional-arguments
def _estimate_kafka_cluster_zonal(  # noqa: C901
    instance: Instance,
    drive: Drive,
    desires: CapacityDesires,
    hot_retention_seconds: float,
    zones_per_region: int = 3,
    copies_per_region: int = 2,
    require_local_disks: bool = False,
    require_attached_disks: bool = False,
    required_zone_size: Optional[int] = None,
    max_regional_size: int = 150,
    max_local_data_per_node_gib: int = 2 * 1024,
    max_attached_data_per_node_gib: int = 2 * 1024,
    min_instance_cpu: int = 2,
    min_instance_memory_gib: int = 12,
    require_same_instance_family: bool = True,
) -> Optional[CapacityPlan]:
    # Kafka doesn't like to deploy on single CPU instances or with < 12 GiB of ram
    if instance.cpu < min_instance_cpu or instance.ram_gib < min_instance_memory_gib:
        return None

    # if we're not allowed to use attached disks, skip EBS only types
    if instance.drive is None and require_local_disks:
        return None

    # if we're not allowed to use local disks, skip ephems
    if instance.drive is not None and require_attached_disks:
        return None

    # Kafka only deploys on gp3 drives right now
    if instance.drive is None and drive.name != "gp3":
        return None

    # If there is a current cluster, check if we are restricted to same instance family
    current_zonal_cluster = _get_current_zonal_cluster(desires)
    if (
        current_zonal_cluster
        and require_same_instance_family
        and not _is_same_instance_family(current_zonal_cluster, instance.family)
    ):
        return None

    requirement, regrets = _estimate_kafka_requirement(
        instance=instance,
        desires=desires,
        zones_per_region=zones_per_region,
        copies_per_region=copies_per_region,
        hot_retention_seconds=hot_retention_seconds,
    )

    # Account for sidecars and base system memory
    base_mem = (
        desires.data_shape.reserved_instance_app_mem_gib
        + desires.data_shape.reserved_instance_system_mem_gib
    )

    # Kafka clusters in prod (tier 0+1) need at least 2 nodes per zone
    min_count_for_tier = 1
    if desires.service_tier < 2:
        min_count_for_tier = 2

    # Kafka read io / second is zonal
    normalized_to_mib = desires.model_copy(deep=True)
    read_mib_per_second: int = (
        int(normalized_to_mib.query_pattern.estimated_mean_read_size_bytes.mid)
        // MIB_IN_BYTES
        // zones_per_region
    )
    write_mib_per_second: int = (
        int(normalized_to_mib.query_pattern.estimated_mean_write_size_bytes.mid)
        // MIB_IN_BYTES
        // zones_per_region
    )

    # All Kafka IOs are sequential, so they can use the group size
    read_ios_per_second = max(1, (read_mib_per_second * 1024) // drive.seq_io_size_kib)
    write_ios_per_second = max(
        1, (write_mib_per_second * 1024) // drive.seq_io_size_kib
    )

    needed_disk_gib = int(requirement.disk_gib.mid)
    disk_buffer_ratio = buffer_for_components(
        buffers=desires.buffers, components=[BufferComponent.disk]
    ).ratio
    max_disk_per_node_gib = get_effective_disk_per_node_gib(
        instance,
        drive,
        disk_buffer_ratio,
        max_local_data_per_node_gib=max_local_data_per_node_gib,
        max_attached_data_per_node_gib=max_attached_data_per_node_gib,
    )
    min_count_for_data = math.ceil(needed_disk_gib / max_disk_per_node_gib)
    cluster = compute_stateful_zone(
        instance=instance,
        drive=drive,
        needed_cores=int(requirement.cpu_cores.mid),
        needed_disk_gib=int(requirement.disk_gib.mid),
        needed_memory_gib=int(requirement.mem_gib.mid),
        needed_network_mbps=requirement.network_mbps.mid,
        # Leave 2x overhead for both reads and writes.
        # Readers are sequential, 100MiB/s read is 100Mib / block_size ios
        # Writers are sequential, 100MiB/s write is 100MiB / block_size ios
        required_disk_ios=lambda size, count: (
            _kafka_read_io(
                rps=read_ios_per_second / count,
                # Kafka does sequential IO
                io_size_kib=drive.seq_io_size_kib,
                size_gib=size,
                # Enough IO to recover a node in 60 minutes
                recovery_seconds=60 * 60,
            ),
            # Leave 100% IO headroom for writes
            copies_per_region * (write_ios_per_second / count) * 2,
        ),
        cluster_size=lambda x: x,
        min_count=max(min_count_for_tier, min_count_for_data, required_zone_size or 1),
        # Sidecars and Variable OS Memory
        # Kafka currently uses 8GiB fixed, might want to change to min(30, x // 2)
        reserve_memory=lambda instance_mem_gib: base_mem + 8,
    )

    # Communicate to the actual provision that if we want reduced RF
    params = {"kafka.copies": copies_per_region}
    _upsert_params(cluster, params)

    # This is roughly the disk we would have tried to provision with the current
    # cluster's instance count (or required_zone_size)
    # If we *could* have satisified the required_zone_size without exceeding the
    # max data size per node constraints but the actual cluster count does not
    # match the required_zone_size, then we want to omit the plan
    # However if there was no way to satisfy the required_zone_size with the
    # max data size per node constraints, then we would rather return some result
    # with a higher instance count since this is the best we can do
    if (
        required_zone_size is not None
        and min_count_for_data <= required_zone_size
        and cluster.count != required_zone_size
    ):
        return None

    # Kafka clusters generally should try to stay under some total number
    # of nodes. Orgs do this for all kinds of reasons such as
    #   * Security group limits. Since you must have < 500 rules if you're
    #       ingressing public ips)
    #   * Maintenance. If your restart script does one node at a time you want
    #       smaller clusters so your restarts don't take months.
    #   * NxN network issues. Sometimes smaller clusters of bigger nodes
    #       are better for network propagation
    if cluster.count > (max_regional_size // zones_per_region):
        return None

    ec2_cost = zones_per_region * cluster.annual_cost

    # Account for the clusters and replication costs
    kafka_costs = {"kafka.zonal-clusters": ec2_cost}

    cluster.cluster_type = "kafka"
    clusters = Clusters(
        annual_costs=kafka_costs,
        zonal=[cluster] * zones_per_region,
        regional=[],
        services=[],
    )

    return CapacityPlan(
        requirements=Requirements(
            zonal=[requirement] * zones_per_region, regrets=regrets
        ),
        candidate_clusters=clusters,
    )


class NflxKafkaArguments(BaseModel):
    copies_per_region: int = Field(
        default=2,
        description=(
            "How many copies of the data will exist e.g. RF=3. If not supplied"
            " this will be determined from mode"
        ),
    )
    cluster_type: ClusterType = Field(
        default=ClusterType.ha,
        description="If the cluster is 'strong' consistency or 'high availability'",
    )
    retention: str = Field(
        default="PT8H", description="How long to retain data in this cluster."
    )
    hot_retention: str = Field(
        default="PT10M",
        description=(
            "How long to retain data in page cache for consumers to skip IO. "
            "Typically consumers lag under 10s, but ensure up to 10M can be handled"
        ),
    )
    require_local_disks: bool = Field(
        default=False,
        description="If local (ephemeral) drives are required",
    )
    require_attached_disks: bool = Field(
        default=False,
        description="If attached (ebs) drives are required",
    )
    required_zonal_size: Optional[int] = Field(
        default=None,
        description="Require zonal clusters to be this size (force vertical scaling)",
    )
    max_regional_size: int = Field(
        default=150,
        description="What is the maximum size of a cluster in this region",
    )
    max_local_disk_gib: int = Field(
        default=5120,
        description=(
            "The maximum amount of data we store per node. Used to limit "
            "recovery duration on failure."
        ),
    )
    min_instance_cpu: int = Field(
        default=2,
        description="The minimum number of instance CPU to allow",
    )
    min_instance_memory_gib: int = Field(
        default=12,
        description="The minimum amount of instance memory to allow",
    )


class NflxKafkaCapacityModel(CapacityModel):
    HA_DEFAULT_REPLICATION_FACTOR = 2
    SC_DEFAULT_REPLICATION_FACTOR = 3

    @staticmethod
    def capacity_plan(
        instance: Instance,
        drive: Drive,
        context: RegionContext,
        desires: CapacityDesires,
        extra_model_arguments: Dict[str, Any],
    ) -> Optional[CapacityPlan]:
        cluster_type: ClusterType = ClusterType(
            extra_model_arguments.get("cluster_type", "high-availability")
        )
        replication_factor = NflxKafkaCapacityModel.HA_DEFAULT_REPLICATION_FACTOR
        if cluster_type == ClusterType.strong:
            replication_factor = NflxKafkaCapacityModel.SC_DEFAULT_REPLICATION_FACTOR
        copies_per_region: int = extra_model_arguments.get(
            "copies_per_region", replication_factor
        )
        if cluster_type == ClusterType.strong and copies_per_region < 3:
            raise ValueError("Strong consistency and RF<3 doesn't work")

        max_regional_size: int = extra_model_arguments.get("max_regional_size", 150)
        # Very large nodes are hard to cache warm
        max_local_data_per_node_gib: int = extra_model_arguments.get(
            "max_local_data_per_node_gib",
            extra_model_arguments.get("max_local_disk_gib", 1024 * 2),
        )
        min_instance_cpu: int = extra_model_arguments.get("min_instance_cpu", 2)
        min_instance_memory_gib: int = extra_model_arguments.get(
            "min_instance_memory_gib", 12
        )
        hot_retention_seconds: float = iso_to_seconds(
            extra_model_arguments.get("hot_retention", "PT10M")
        )
        require_local_disks: bool = extra_model_arguments.get(
            "require_local_disks", False
        )
        require_attached_disks: bool = extra_model_arguments.get(
            "require_attached_disks", False
        )
        required_zone_size: Optional[int] = extra_model_arguments.get(
            "required_zone_size", None
        )
        # By default, for existing clusters, restrict to only using same instance family
        require_same_instance_family: bool = extra_model_arguments.get(
            "require_same_instance_family", True
        )

        return _estimate_kafka_cluster_zonal(
            instance=instance,
            drive=drive,
            desires=desires,
            zones_per_region=context.zones_in_region,
            copies_per_region=copies_per_region,
            require_local_disks=require_local_disks,
            require_attached_disks=require_attached_disks,
            required_zone_size=required_zone_size,
            max_regional_size=max_regional_size,
            max_local_data_per_node_gib=max_local_data_per_node_gib,
            min_instance_cpu=min_instance_cpu,
            min_instance_memory_gib=min_instance_memory_gib,
            hot_retention_seconds=hot_retention_seconds,
            require_same_instance_family=require_same_instance_family,
        )

    @staticmethod
    def description() -> str:
        return "Netflix Streaming Kafka Model"

    @staticmethod
    def extra_model_arguments_schema() -> Dict[str, Any]:
        return NflxKafkaArguments.model_json_schema()

    @staticmethod
    def allowed_cloud_drives() -> Tuple[Optional[str], ...]:
        # Kafka at Netflix only supports 3rd gen SSD storage
        return ("gp3",)

    @staticmethod
    def default_desires(
        user_desires: CapacityDesires, extra_model_arguments: Dict[str, Any]
    ) -> CapacityDesires:
        # Default to 10MiB/s and a single reader
        concurrent_readers = max(
            1, int(user_desires.query_pattern.estimated_read_per_second.mid)
        )
        query_pattern = user_desires.query_pattern.model_dump()
        if "estimated_mean_write_size_bytes" in query_pattern:
            write_bytes = Interval(**query_pattern["estimated_mean_write_size_bytes"])
        else:
            write_bytes = certain_int(10 * 1024 * 1024)

        read_bytes = write_bytes.scale(concurrent_readers)
        retention = extra_model_arguments.get("retention", "PT8H")
        retention_secs = iso_to_seconds(retention)

        # write throughput * retention * replication factor = usage
        replication_factor = NflxKafkaCapacityModel.HA_DEFAULT_REPLICATION_FACTOR
        if extra_model_arguments.get("cluster_type", None) == ClusterType.strong:
            replication_factor = NflxKafkaCapacityModel.SC_DEFAULT_REPLICATION_FACTOR
        state_gib = (
            write_bytes.mid * retention_secs * replication_factor
        ) / GIB_IN_BYTES

        # By supplying these buffers we can deconstruct observed utilization into
        # load versus buffer.
        compute_buffer_ratio = 2.5 if user_desires.service_tier in (0, 1) else 2.0
        buffers = Buffers(
            default=Buffer(ratio=1.5),
            desired={
                # Amount of compute buffer that we need to reserve in addition to
                # cpu_headroom_target that is reserved on a per instance basis
                "compute": Buffer(
                    ratio=compute_buffer_ratio, components=[BufferComponent.compute]
                ),
                # This makes sure we use only 40% of the available storage
                "storage": Buffer(ratio=2.5, components=[BufferComponent.storage]),
            },
        )

        return CapacityDesires(
            query_pattern=QueryPattern(
                access_pattern=AccessPattern.throughput,
                access_consistency=GlobalConsistency(
                    same_region=Consistency(
                        target_consistency=AccessConsistency.read_your_writes
                    ),
                    cross_region=Consistency(
                        target_consistency=AccessConsistency.never
                    ),
                ),
                estimated_mean_write_size_bytes=write_bytes,
                estimated_mean_read_size_bytes=read_bytes,
                # How much on-cpu time to read or write 1 MiB
                # 42 MiB / second in 3 CPU seconds = 14 MiB / CPU second
                # This = 1000 / 14 = 71 millisecond / MiB
                # To read 1 MiB costs 71 millisecond of CPU time
                # Later tuned down based on training data
                estimated_mean_read_latency_ms=Interval(
                    low=20, mid=40, high=75, confidence=0.98
                ),
                # Wild-Ass-Guess that writes are slightly cheaper
                # Because no disks/context switching??
                estimated_mean_write_latency_ms=Interval(
                    low=20, mid=30, high=75, confidence=0.98
                ),
            ),
            data_shape=DataShape(
                estimated_state_size_gib=Interval(
                    low=state_gib * 0.5,
                    mid=state_gib,
                    high=state_gib * 2,
                    confidence=0.98,
                ),
                # Gandalf, system stuffs
                reserved_instance_app_mem_gib=1,
                # Connection overhead, kernel, etc ...
                reserved_instance_system_mem_gib=3,
            ),
            buffers=buffers,
        )


nflx_kafka_capacity_model = NflxKafkaCapacityModel()



import math
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple

from service_capacity_modeling.models import CapacityPlan


def reduce_by_family(
    plans: Iterable[CapacityPlan], max_results_per_family: int = 1
) -> List[CapacityPlan]:
    """Groups a potential set of clusters by hardware family sorted by cost.

    Useful for showing different family options.

    Args:
        plans: Iterable of CapacityPlan objects to filter
        max_results_per_family: Maximum number of results to return per
            family combination
    """
    zonal_families: Dict[Tuple[Tuple[str, str], ...], int] = {}
    regional_families: Dict[Tuple[Tuple[str, str], ...], int] = {}

    result: List[CapacityPlan] = []
    for plan in plans:
        topo = plan.candidate_clusters
        regional_type: Tuple[Tuple[str, str], ...] = tuple()
        zonal_type: Tuple[Tuple[str, str], ...] = tuple()

        if topo.regional:
            regional_type = tuple(
                sorted({(c.cluster_type, c.instance.family) for c in topo.regional})
            )

        if topo.zonal:
            zonal_type = tuple(
                sorted({(c.cluster_type, c.instance.family) for c in topo.zonal})
            )

        # Count how many of each family combination we've seen
        zonal_count = zonal_families.get(zonal_type, 0)
        regional_count = regional_families.get(regional_type, 0)

        # Add the plan if we haven't reached the maximum for either family type
        if (
            zonal_count < max_results_per_family
            or regional_count < max_results_per_family
        ):
            result.append(plan)

            # Update counters
            zonal_families[zonal_type] = zonal_count + 1
            regional_families[regional_type] = regional_count + 1

    return result


# https://stackoverflow.com/questions/14267555/find-the-smallest-power-of-2-greater-than-or-equal-to-n-in-python
def next_power_of_2(y: float) -> int:
    x = int(y)
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


def is_power_of_2(y: float) -> bool:
    """Check if x is a power of 2 or 1"""
    return y == next_power_of_2(y)


def next_doubling(x: float, base: int) -> int:
    # Some clusters were provisioned as non powers of (e.g. 12)
    # And so if a requirement cannot be satisifed by a cluster
    # we would want to round up to the next doubling
    # E.g. 12 -> 24 -> 48
    # e.g. 3 -> 6 -> 12
    if x <= base:
        return int(base)
    return int(base * (2 ** math.ceil(math.log(x / base, 2))))


def next_n(x: float, n: float) -> int:
    return int(math.ceil(x / n)) * int(n)



import setuptools

setuptools.setup(
    name="service-capacity-modeling",
    author="Joseph Lynch",
    author_email="josephl@netflix.com",
    versioning="distance",
    setup_requires="setupmeta",
    description="Contains utilities for modeling capacity for pluggable workloads",
    python_requires=">=3.10,<3.13",
    packages=setuptools.find_packages(exclude=("tests*", "notebooks*")),
    install_requires=[
        "pydantic>2.0",
        "scipy",
        "numpy",
        "isodate",
    ],
    extras_require={
        "aws": ["boto3"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "auto-shape = service_capacity_modeling.tools.auto_shape:main",
            "fetch-pricing = service_capacity_modeling.tools.fetch_pricing:main",
        ]
    },
    include_package_data=True,
    package_data={
        "": [
            "hardware/profiles/profiles.txt",
            "hardware/profiles/shapes/**/*.json",
            "hardware/profiles/pricing/**/*.json",
        ]
    },
)



import argparse
import json
import re
import sys
from fractions import Fraction
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple

import boto3
import botocore
from pydantic import BaseModel

from service_capacity_modeling.hardware import shapes
from service_capacity_modeling.interface import Drive
from service_capacity_modeling.interface import DriveType
from service_capacity_modeling.interface import FixedInterval
from service_capacity_modeling.interface import Instance
from service_capacity_modeling.interface import normalized_aws_size


# Default latency curves from FIO testing per generation
latency_curve_ms: Dict[str, FixedInterval] = {
    "7th-gen-ephemeral": FixedInterval(
        low=0.071,
        mid=0.079,
        high=0.258,
        confidence=0.9,
        minimum_value=0.026,
        maximum_value=2.153,
    ),
    "6th-gen-ssd": FixedInterval(
        low=0.1,
        mid=0.125,
        high=0.170,
        confidence=0.9,
        minimum_value=0.050,
        maximum_value=2.000,
    ),
    "5th-gen-ssd": FixedInterval(
        low=0.080,
        mid=0.125,
        high=0.200,
        confidence=0.9,
        minimum_value=0.070,
        maximum_value=2.000,
    ),
    "ssd": FixedInterval(low=0.08, mid=0.10, high=0.2),
    "hdd": FixedInterval(low=1, mid=2, high=10),
}

# Default xlarge -> (4k random read, write) iops tables
aws_xlarge_iops = {
    # General Purpose and Memory Share IOPs
    "5ad": (59_000, 29_000),
    "5d": (59_000, 29_000),
    "5dn": (58_000, 29_000),
    "6gd": (53_750, 22_500),
    "6id": (67_083, 33_542),
    "6idn": (67_083, 33_542),
    "7gd": (67_083, 33_542),
    # Compute has smaller IOPs
    "c5ad": (32_566, 14_211),
    "c5d": (40_000, 18_000),
    "c6gd": (53_750, 22_500),
    "c6id": (67_083, 33_542),
    "c7gd": (67_083, 33_542),
    # Storage has more
    # https://docs.aws.amazon.com/ec2/latest/instancetypes/so.html#so_instance-store
    "i3": (206_250, 70_000),
    "i3en": (85_000, 65_000),
    "i4g": (62_500, 50_000),
    "i4i": (100_000, 55_000),
    "i7i": (150_000, 82_500),
    "i7ie": (108_333, 86_666),
    "i8g": (150_000, 82_500),
}


def guess_iops(family: str) -> Optional[Tuple[int, int]]:
    if family[0] in ("m", "r"):
        return aws_xlarge_iops.get(family[1:])
    return aws_xlarge_iops.get(family)


aws_io_links = {
    "m": "https://docs.aws.amazon.com/ec2/latest/instancetypes/gp.html#gp_instance-store",
    "c": "https://docs.aws.amazon.com/ec2/latest/instancetypes/co.html#co_instance-store",
    "r": "https://docs.aws.amazon.com/ec2/latest/instancetypes/mo.html#mo_instance-store",
    "i": "https://docs.aws.amazon.com/ec2/latest/instancetypes/so.html#so_instance-store",
}
aws_instance_link = "https://docs.aws.amazon.com/ec2/latest/instancetypes/ec2-instance-type-specifications.html"


class CPUPerformance(BaseModel):
    ipc_scale_factor: Optional[float] = None


class IOPerformance(BaseModel):
    xl_read_iops: float
    xl_write_iops: float
    latency: FixedInterval
    single_tenant_size: int = 0


def _gb_to_gib(inp: float) -> int:
    return int(round((inp * 10**9) / 2.0**30))


def _drive(
    drive_type: DriveType,
    io_perf: Optional[IOPerformance],
    scale: Fraction,
    data: Dict[str, Any],
) -> Optional[Drive]:
    if drive_type.name.startswith("attached") or io_perf is None:
        return None

    size = data["InstanceStorageInfo"]["TotalSizeInGB"]
    size_gib = _gb_to_gib(size)
    single_drive_size = data["InstanceStorageInfo"]["Disks"][0]["SizeInGB"]
    single_tenant = False
    if single_drive_size == io_perf.single_tenant_size:
        single_tenant = True

    return Drive(
        name="ephem",
        size_gib=int(size_gib),
        write_io_per_s=int(round(io_perf.xl_write_iops * scale)),
        read_io_per_s=int(round(io_perf.xl_read_iops * scale)),
        read_io_latency_ms=io_perf.latency,
        single_tenant=single_tenant,
    )


def pull_family(
    ec2_client: Any,
    family: str,
    cpu_perf: Optional[CPUPerformance] = None,
    io_perf: Optional[IOPerformance] = None,
    debug: bool = False,
) -> Sequence[Instance]:
    # flake8: noqa: C901
    """
    Pull Instance shapes from AWS APIs.

    Emulates something like the following

    aws --region us-east-1 \
      ec2 describe-instance-types \
      --filters 'Name=instance-type,Values=m7a.*'

    And then convert from base 10 for memory and disk to base 2 and generate the
    shape from that.

    Returns a list of Instance for hardware. Then we can compose to writing
    json files.
    """
    request = ec2_client.get_paginator("describe_instance_types")
    payload = {
        "Filters": [
            {
                "Name": "instance-type",
                "Values": [
                    family + ".*",
                ],
            },
        ],
    }

    def debug_log(msg: str) -> None:
        if debug:
            print(msg, file=sys.stderr)

    instance_jsons_dict = {}
    disk_type = DriveType.attached_ssd
    disk_max_size = -1
    for instance_page in request.paginate(**payload):
        for instance_type_json in instance_page["InstanceTypes"]:
            debug_log(json.dumps(instance_type_json))
            if "metal" in instance_type_json["InstanceType"]:
                debug_log("Excluding metal for now.")
                continue

            if instance_type_json["InstanceStorageSupported"] is True:
                typ = instance_type_json["InstanceStorageInfo"]["Disks"][0]["Type"]
                disk_type = DriveType["local_" + typ.lower()]
                for disk in instance_type_json["InstanceStorageInfo"]["Disks"]:
                    # The biggest size drive is the single tenant one
                    disk_max_size = max(disk_max_size, int(disk["SizeInGB"]))

            instance_jsons_dict[instance_type_json["InstanceType"].split(".")[1]] = (
                instance_type_json
            )

    if disk_type.name.startswith("local"):
        if (
            io_perf is None
            or io_perf.xl_read_iops is None
            or io_perf.xl_write_iops is None
        ):
            link = aws_io_links.get(family[0], aws_instance_link)
            print(
                "Instance shape has ephemeral storage. You must pass --xl-iops with "
                "data either from fio benchmarking or from AWS's page for the "
                "appropriate family:\n"
                f"Search for {family}.xlarge in {link}\n",
                file=sys.stderr,
            )
            sys.exit(2)
        debug_log(
            f"{disk_type.name} IO data:\n"
            f"xlarge  read IOPS: {io_perf.xl_read_iops}\n"
            f"xlarge write IOPS: {io_perf.xl_write_iops}\n"
            f"latency curve: {io_perf.latency.model_dump_json()}\n"
        )
        io_perf = io_perf.model_copy(
            update={
                "single_tenant_size": disk_max_size,
            }
        )

    results = []
    # Now build the instance shapes from the data
    for _, data in instance_jsons_dict.items():
        name = data["InstanceType"]
        try:
            normalized_size = normalized_aws_size(name)
        except AssertionError:
            print(name)

        drive = _drive(disk_type, io_perf, scale=normalized_size, data=data)
        if cpu_perf is not None and cpu_perf.ipc_scale_factor is not None:
            cpu_ipc_scale_factor = cpu_perf.ipc_scale_factor
        else:
            if data["VCpuInfo"].get("DefaultThreadsPerCore", 2) == 1:
                cpu_ipc_scale_factor = 1.5
            else:
                cpu_ipc_scale_factor = 1.0

        new_shape = Instance(
            name=data["InstanceType"],
            cpu=data["VCpuInfo"]["DefaultVCpus"],
            cpu_cores=data["VCpuInfo"]["DefaultCores"],
            cpu_ghz=data["ProcessorInfo"]["SustainedClockSpeedInGhz"],
            cpu_ipc_scale=cpu_ipc_scale_factor,
            ram_gib=round(data["MemoryInfo"]["SizeInMiB"] * 10**6 / 2.0**30, 2),
            net_mbps=round(
                data["NetworkInfo"]["NetworkCards"][0]["BaselineBandwidthInGbps"]
                * 10**3
            ),
            drive=drive,
        )

        results.append(new_shape)
    results = sorted(results, key=lambda i: normalized_aws_size(i.name))
    return results


def parse_iops(inp: Optional[str]) -> Optional[Tuple[int, int]]:
    """Parses strings like 100,000/50,000 to (100000, 50000)"""
    if inp is None:
        return None

    # AWS often gives like 117,000 / 57,000
    if inp.count("/") == 1:
        left, right = inp.split("/")
        left = left.strip().replace(",", "")
        right = right.strip().replace(",", "")
        return int(left), int(right)
    else:
        raise argparse.ArgumentTypeError(
            "xlarge IOPS should be given in <random read>/<write> format. "
            "For example r5d.xlarge would be 59000/29000."
        )


def _parse_family(family: str) -> Tuple[str, int, str]:
    series = family[0]
    num = re.findall(r"\d+", family)
    assert len(num) == 1
    gen = int(num[0])

    vendor = "intel"
    if family[-1] == "a":
        vendor = "amd"
    if family[-1] == "g":
        vendor = "graviton"

    return series, gen, vendor


def parse_io_curve(inp: str, family: str) -> FixedInterval:
    series, gen, _ = _parse_family(family)

    if series in ("i", "d") and gen < 7:
        # i4i is actually 6th gen
        # i3/i3en/d3 is actually 5th gen
        # AWS fixed this after 7th gen
        gen += 2

    if f"{gen}th-gen-{inp}" in latency_curve_ms:
        return latency_curve_ms[f"{gen}th-gen-{inp}"]

    return latency_curve_ms[inp]


def deduce_io_perf(
    family: str, curve: str, iops: Optional[Tuple[int, int]]
) -> Optional[IOPerformance]:
    if iops is None:
        guess = guess_iops(family)
        if guess is None:
            return None
        else:
            return IOPerformance(
                latency=parse_io_curve(curve, family),
                xl_read_iops=guess[0],
                xl_write_iops=guess[1],
            )
    else:
        return IOPerformance(
            latency=parse_io_curve(curve, family),
            xl_read_iops=iops[0],
            xl_write_iops=iops[1],
        )


def deduce_cpu_perf(
    family: str, ipc_scale_factor: Optional[float] = None
) -> Optional[CPUPerformance]:
    if ipc_scale_factor is not None:
        return CPUPerformance(
            ipc_scale_factor=ipc_scale_factor,
        )
    else:
        _, gen, vendor = _parse_family(family)
        if vendor == "intel" and gen == 7:
            # The 7th generation intel machines have a lower base clock
            # But in real world workloads are around 15% faster. So
            # (1.15 * 3.5) / 3.2 = 1.25
            return CPUPerformance(ipc_scale_factor=1.25)
    return CPUPerformance()


def main(args: Any) -> int:
    for family in args.families:
        io_perf = deduce_io_perf(
            family=family, curve=args.io_latency_curve, iops=args.xl_iops
        )

        cpu_perf = deduce_cpu_perf(
            family=family,
            ipc_scale_factor=args.cpu_ipc_scale,
        )

        ec2_client = boto3.client("ec2", region_name=args.region)
        family_shapes = pull_family(
            ec2_client=ec2_client,
            family=family,
            cpu_perf=cpu_perf,
            io_perf=io_perf,
            debug=args.debug,
        )
        json_shapes: Dict[str, Dict[str, Any]] = {"instances": {}}
        for shape in family_shapes:
            model_dict = shape.model_dump(exclude_unset=True)
            # Hardware shouldn't have costs
            if "annual_cost" in model_dict:
                del model_dict["annual_cost"]
            if (
                "drive" in model_dict
                and model_dict["drive"] is not None
                and "annual_cost" in model_dict["drive"]
            ):
                del model_dict["drive"]["annual_cost"]

            json_shapes["instances"][shape.name] = model_dict

        print(f"[{family}] Hardware Shapes", file=sys.stderr)
        print(json.dumps(json_shapes, indent=2))

        # Write to JSON file if requested
        if args.output_path is not None:
            path: Path = args.output_path
            if path.is_dir():
                output_path = Path(path, f"auto_{family}.json")
            else:
                output_path = path
            with open(output_path, "wt", encoding="utf-8") as fd:
                json.dump(json_shapes, fd, indent=2)
                fd.write("\n")

    return 0


if __name__ == "__main__":
    families = set()
    regions = set()
    for ar in shapes.hardware.regions.keys():
        regions.add(ar)

    try:
        client = boto3.client("ec2", region_name="us-east-1")
        result = client.describe_instance_type_offerings()
    except botocore.exceptions.ClientError:
        print(
            "Unable to connect to EC2. Do you have AWS credentials refreshed?",
            file=sys.stderr,
        )
        sys.exit(1)

    offerings = result["InstanceTypeOfferings"]
    for offering in offerings:
        families.add(offering["InstanceType"].rsplit(".", 1)[0])

    parser = argparse.ArgumentParser(
        prog="Project shapes from instance family filter",
        description="Input the target instance family filter like m7a.* and generates shape data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--xl-iops",
        default=None,
        type=parse_iops,
        help=(
            "The xlarge size's iops expressed as <rand 4k reads>/<write>. "
            "For example '100,000 / 50,000'. To find this information use AWS's "
            f"spec: {aws_instance_link} OR use fio to benchmark. Deduced from family "
            " if we can"
        ),
    )
    parser.add_argument(
        "--io-latency-curve", default="ssd", choices=latency_curve_ms.keys()
    )
    parser.add_argument(
        "--cpu-ipc-scale",
        type=float,
        help=(
            "If the clock frequency is not a good measure of performance, scale it by "
            "this much. Note that by default this will be 1, unless the cores are all "
            "full cores (not threads) in which case it will be 1.5 by default."
        ),
    )
    parser.add_argument("--region", choices=regions, default="us-east-1")
    parser.add_argument(
        "--output-path",
        type=Path,
        help=(
            "Output file path to copy the result to. If not given only stdout will"
            "occur. If running from the repo use: "
            "service_capacity_modeling/hardware/profiles/shapes/aws/"
        ),
    )
    parser.add_argument("--debug", action="store_true", help="Show verbose output")

    parser.add_argument(
        "families",
        nargs="+",
        type=str,
        default="m7a",
        metavar="family",
        choices=sorted(list(families)),
    )
    sys.exit(main(parser.parse_args()))



from typing import List
from unittest.mock import Mock

from service_capacity_modeling.capacity_planner import CapacityPlanner
from service_capacity_modeling.interface import CapacityDesires
from service_capacity_modeling.interface import certain_float
from service_capacity_modeling.interface import certain_int
from service_capacity_modeling.interface import CurrentClusters
from service_capacity_modeling.interface import CurrentRegionClusterCapacity
from service_capacity_modeling.interface import Lifecycle
from service_capacity_modeling.interface import Platform


def test_generate_scenarios():
    # Create a CapacityPlanner instance
    planner = CapacityPlanner()

    # Mock the model
    model = Mock()
    model.allowed_platforms.return_value = {Platform.amd64}
    model.allowed_cloud_drives.return_value = {}
    model.run_hardware_simulation.return_value = True

    region = "us-east-1"
    desires = CapacityDesires()
    num_regions = 3
    lifecycles = [Lifecycle.stable]
    instance_families: List[str] = []
    drives: List[str] = []

    # Call generate_scenarios
    scenarios = list(
        planner.generate_scenarios(
            model, region, desires, num_regions, lifecycles, instance_families, drives
        )
    )

    # Check we got some instances
    assert len(scenarios) > 0


def test_generate_scenarios_limit_family():
    # Create a CapacityPlanner instance
    planner = CapacityPlanner()

    # Mock the model
    model = Mock()
    model.allowed_platforms.return_value = {Platform.amd64}
    model.allowed_cloud_drives.return_value = {}
    model.run_hardware_simulation.return_value = True

    region = "us-east-1"
    desires = CapacityDesires()
    num_regions = 3
    lifecycles = [Lifecycle.stable]
    # Limit to m5 family
    instance_families = ["m5"]
    drives: List[str] = []

    # Call generate_scenarios
    scenarios = list(
        planner.generate_scenarios(
            model, region, desires, num_regions, lifecycles, instance_families, drives
        )
    )

    # Check we got m5s
    assert len(scenarios) > 0
    for instance, _, _ in scenarios:
        assert instance.family == "m5"


def test_generate_scenarios_desire_resources():
    # Create a CapacityPlanner instance
    planner = CapacityPlanner()

    # Mock the model
    model = Mock()
    model.allowed_platforms.return_value = {Platform.amd64}
    model.allowed_cloud_drives.return_value = {}
    model.run_hardware_simulation.return_value = True

    region = "us-east-1"
    num_regions = 3
    lifecycles = [Lifecycle.stable]
    # Limit to m5 family
    instance_families = ["m5"]
    drives: List[str] = []

    desires = CapacityDesires()
    # Set cpu and memory requirements via application desires
    desires.data_shape.reserved_instance_app_mem_gib = 64
    desires.data_shape.reserved_instance_system_mem_gib = 0
    desires.query_pattern.estimated_read_parallelism = certain_int(16)
    desires.query_pattern.estimated_write_parallelism = certain_int(0)

    # Call generate_scenarios
    scenarios = list(
        planner.generate_scenarios(
            model, region, desires, num_regions, lifecycles, instance_families, drives
        )
    )

    # Check we got m5s
    assert len(scenarios) > 0
    for instance, _, _ in scenarios:
        assert instance.family == "m5"
        assert instance.cpu >= 16
        assert instance.ram_gib >= 64


def test_generate_scenarios_current_resources():
    # Create a CapacityPlanner instance
    planner = CapacityPlanner()

    # Mock the model
    model = Mock()
    model.allowed_platforms.return_value = {Platform.amd64}
    model.allowed_cloud_drives.return_value = {}
    model.run_hardware_simulation.return_value = True

    region = "us-east-1"
    num_regions = 3
    lifecycles = [Lifecycle.stable]
    # Limit to m5 family
    instance_families = ["m5"]
    drives: List[str] = []

    desires = CapacityDesires()
    # Set cpu and memory requirements via current capacity
    desires.current_clusters = CurrentClusters()
    desires.current_clusters.regional = [
        CurrentRegionClusterCapacity(
            cluster_instance_name="m5.12xlarge",
            cluster_instance_count=certain_int(5),
            cpu_utilization=certain_float(100.0),
            memory_utilization_gib=certain_float(16),
            network_utilization_mbps=certain_float(128.0),
        )
    ]

    # Call generate_scenarios
    scenarios = list(
        planner.generate_scenarios(
            model, region, desires, num_regions, lifecycles, instance_families, drives
        )
    )

    # Check we got m5s
    assert len(scenarios) > 0
    for instance, _, _ in scenarios:
        assert instance.family == "m5"
        assert instance.ram_gib >= 16
